# -*- coding: utf-8 -*-
"""
Unittests for uxding module
"""
import sys
import unittest

import os
import tempfile
import shutil

from ioflo.aid.sixing import *
from ioflo.aid.consoling import getConsole
from ioflo.aio import wiring
from ioflo.aio.win import wining

console = getConsole()

def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    console.reinit(verbosity=console.Wordage.concise)

class BasicTestCase(unittest.TestCase):
    """
    Test Case
    """

    def setUp(self):
        """

        """
        pass

    def tearDown(self):
        """

        """
        pass


    def testWinMailslotNb(self):
        """
        Test Class SocketUxdNb
        """
        console.terse("{0}\n".format(self.testWinMailslotNb.__doc__))
        console.reinit(verbosity=console.Wordage.profuse)

        try:
            import win32file
        except ImportError:
            return

        userDirpath = os.path.join('~', '.ioflo', 'test')
        userDirpath = os.path.abspath(os.path.expanduser(userDirpath))
        if not os.path.exists(userDirpath):
            os.makedirs(userDirpath)

        tempDirpath = tempfile.mkdtemp(prefix="test", suffix="wms", dir=userDirpath)

        sockDirpath = os.path.join(tempDirpath, 'wms')
        if not os.path.exists(sockDirpath):
            os.makedirs(sockDirpath)

        logDirpath = os.path.join(tempDirpath, 'log')
        if not os.path.exists(logDirpath):
            os.makedirs(logDirpath)

        wireLog = wiring.WireLog(path=logDirpath)
        result = wireLog.reopen(prefix='alpha', midfix='wms')


        ha = os.path.join(sockDirpath, 'alpha.wms')
        alpha = wining.WinMailslotNb(ha=ha, wlog=wireLog)
        result = alpha.reopen()
        self.assertIs(result, True)
        self.assertEqual(alpha.ha, ha)

        ha = os.path.join(sockDirpath, 'beta.wms')
        beta = uxding.SocketUxdNb(ha=ha, umask=0x077)
        result = beta.reopen()
        self.assertIs(result, True)
        self.assertEqual(beta.ha, ha)

        ha = os.path.join(sockDirpath, 'gamma.wms')
        gamma = uxding.SocketUxdNb(ha=ha, umask=0x077)
        result = gamma.reopen()
        self.assertIs(result, True)
        self.assertEqual(gamma.ha, ha)

        txMsg = b"Alpha sends to Beta"
        alpha.send(txMsg, beta.ha)
        rxMsg, sa = beta.receive()
        self.assertEqual(txMsg, rxMsg)
        self.assertEqual(sa, alpha.ha)

        txMsg = b"Alpha sends to Gamma"
        alpha.send(txMsg, gamma.ha)
        rxMsg, sa = gamma.receive()
        self.assertEqual(txMsg, rxMsg)
        self.assertEqual(sa, alpha.ha)

        txMsg = b"Alpha sends to Alpha"
        alpha.send(txMsg, alpha.ha)
        rxMsg, sa = alpha.receive()
        self.assertEqual(txMsg, rxMsg)
        self.assertEqual(sa, alpha.ha)

        txMsg = b"Beta sends to Alpha"
        beta.send(txMsg, alpha.ha)
        rxMsg, sa = alpha.receive()
        self.assertEqual(txMsg, rxMsg)
        self.assertEqual(sa, beta.ha)

        txMsg = b"Beta sends to Gamma"
        beta.send(txMsg, gamma.ha)
        rxMsg, sa = gamma.receive()
        self.assertEqual(txMsg, rxMsg)
        self.assertEqual(sa, beta.ha)

        txMsg = b"Beta sends to Beta"
        beta.send(txMsg, beta.ha)
        rxMsg, sa = beta.receive()
        self.assertEqual(txMsg, rxMsg)
        self.assertEqual(sa, beta.ha)

        txMsg = b"Gamma sends to Alpha"
        gamma.send(txMsg, alpha.ha)
        rxMsg, sa = alpha.receive()
        self.assertEqual(txMsg, rxMsg)
        self.assertEqual(sa, gamma.ha)

        txMsg = b"Gamma sends to Beta"
        gamma.send(txMsg, beta.ha)
        rxMsg, sa = beta.receive()
        self.assertEqual(txMsg, rxMsg)
        self.assertEqual(sa, gamma.ha)

        txMsg = b"Gamma sends to Gamma"
        gamma.send(txMsg, gamma.ha)
        rxMsg, sa = gamma.receive()
        self.assertEqual(txMsg, rxMsg)
        self.assertEqual(sa, gamma.ha)


        pairs = [(alpha, beta), (alpha, gamma), (alpha, alpha),
                 (beta, alpha), (beta, gamma), (beta, beta),
                 (gamma, alpha), (gamma, beta), (gamma, gamma),]
        names = [('alpha', 'beta'), ('alpha', 'gamma'), ('alpha', 'alpha'),
                 ('beta', 'alpha'), ('beta', 'gamma'), ('beta', 'beta'),
                 ('gamma', 'alpha'), ('gamma', 'beta'), ('gamma', 'gamma'),]

        for i, pair in enumerate(pairs):
            txer, rxer = pair
            txName, rxName =  names[i]
            txMsg = ns2b("{0} sends to {1} again".format(txName.capitalize(), rxName.capitalize()))
            txer.send(txMsg, rxer.ha)
            rxMsg, sa = rxer.receive()
            self.assertEqual(txMsg, rxMsg)
            self.assertEqual(sa, txer.ha)


        rxMsg, sa = alpha.receive()
        self.assertIs(b'', rxMsg)
        self.assertIs(None, sa)

        rxMsg, sa = beta.receive()
        self.assertIs(b'', rxMsg)
        self.assertIs(None, sa)

        rxMsg, sa = gamma.receive()
        self.assertIs(b'', rxMsg)
        self.assertIs(None, sa)

        alpha.close()
        beta.close()
        gamma.close()

        wireLog.close()
        shutil.rmtree(tempDirpath)
        console.reinit(verbosity=console.Wordage.concise)


def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
             'testWinMailslotNb',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #console.reinit(verbosity=console.Wordage.concise)

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testClientAutoReconnect')





# -*- coding: utf-8 -*-
"""
Unit Test Template
"""
from __future__ import absolute_import, division, print_function

import sys

import unittest

import os
import array

from ioflo.aid.sixing import *
from ioflo.test import testing
from ioflo.aid.consoling import getConsole


console = getConsole()

from ioflo.aid import vectoring


def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    pass


class BasicTestCase(unittest.TestCase):
    """
    Example TestCase
    """

    def setUp(self):
        """
        Call super if override so House Framer and Frame are setup correctly
        """
        super(BasicTestCase, self).setUp()

    def tearDown(self):
        """
        Call super if override so House Framer and Frame are torn down correctly
        """
        super(BasicTestCase, self).tearDown()

    def testMagnitude(self):
        """
        Test the mag and mag2 functions
        """
        console.terse("{0}\n".format(self.testMagnitude.__doc__))

        from ioflo.aid.vectoring import mag, mag2

        v = [1, 2, 3]  # list
        m = mag(v)
        self.assertAlmostEqual(m, 3.7416573867739413)

        m2 = mag2(v)
        self.assertEqual(m2, 14)

        v = (1, 2, 3)  # tuple
        m = mag(v)
        self.assertAlmostEqual(m, 3.7416573867739413)

        m2 = mag2(v)
        self.assertEqual(m2, 14)

        v = array.array('i', [1, 2, 3])  # array
        m = mag(v)
        self.assertAlmostEqual(m, 3.7416573867739413)

        m2 = mag2(v)
        self.assertEqual(m2, 14)

        v = bytearray([1, 2, 3])
        m = mag(v)
        self.assertAlmostEqual(m, 3.7416573867739413)

        m2 = mag2(v)
        self.assertEqual(m2, 14)

        v = [0, 0, 0]
        m = mag(v)
        self.assertEqual(m, 0.0)

        m2 = mag2(v)
        self.assertEqual(m2, 0.0)

        v = [1, 0, 0]
        m = mag(v)
        self.assertEqual(m, 1.0)

        m2 = mag2(v)
        self.assertEqual(m2, 1)

        v = [-1, -2, -3]  # list
        m = mag(v)
        self.assertAlmostEqual(m, 3.7416573867739413)

        m2 = mag2(v)
        self.assertEqual(m2, 14)

        v = [-1, 0.5, -2.5]  # list
        m = mag(v)
        self.assertAlmostEqual(m, 2.7386127875258306)

        m2 = mag2(v)
        self.assertEqual(m2, 7.5)

    def testAddNegSubMult(self):
        """
        Test the vector add and neg and sub multfunctions
        """
        console.terse("{0}\n".format(self.testAddNegSubMult.__doc__))

        from ioflo.aid.vectoring import add, neg, sub, mult

        u = (1, 2, 3)
        v = (4, 5, 6)
        w = add(u, v)
        self.assertEqual(w, (5, 7, 9))

        w = neg(u)
        self.assertEqual(w, (-1, -2, -3))

        w = add(u, neg(v))
        self.assertEqual(w, (-3, -3, -3))

        w = sub(u, v)
        self.assertEqual(w, (-3, -3, -3))

        w = mult(2, u)
        self.assertEqual(w, (2, 4, 6))

    def testNormalize(self):
        """
        Test the norm function
        """
        console.terse("{0}\n".format(self.testNormalize.__doc__))

        from ioflo.aid.vectoring import norm, mag

        v = [1, 2, 3]  # list
        n = norm(v)
        m = mag(n)
        self.assertAlmostEqual(m, 1.0)

        z = [0.2672612419124244, 0.5345224838248488, 0.8017837257372732]
        for e1, e2 in zip(n, z):
            self.assertAlmostEqual(e1, e2)

        v = [0, 0, 0]  # special case
        n = norm(v)
        m = mag(n)
        self.assertAlmostEqual(m, 0.0)

        v = [1, 0, 0]
        n = norm(v)
        m = mag(n)
        self.assertEqual(m, 1.0)

        z = [1.0, 0.0, 0.0]
        for e1, e2 in zip(n, z):
            self.assertEqual(e1, e2)

        v = [-1, -2, 3]  # list
        n = norm(v)
        m = mag(n)
        self.assertAlmostEqual(m, 1.0)

        z = [-0.2672612419124244, -0.5345224838248488, 0.8017837257372732]
        for e1, e2 in zip(n, z):
            self.assertAlmostEqual(e1, e2)

        v = (0.49999999999999994, 0.8660254037844386)
        n = norm(v, check=True)
        m = mag(n)
        self.assertEqual(m, 1.0)

    def testDotProduct(self):
        """
        Test the dot function
        """
        console.terse("{0}\n".format(self.testDotProduct.__doc__))

        from ioflo.aid.vectoring import dot, norm, mag

        u = [1, 2, 3]  # list
        v = [4, 5, 6]
        m = dot(u, v)
        self.assertEqual(m, 32)
        n = dot(v, u)
        self.assertEqual(m, n)

        v = [0, 0, 0]
        m = dot(u, v)
        self.assertEqual(m, 0.0)

        v = [0, 0, 1]
        m = dot(u, v)
        self.assertEqual(m, 3)
        self.assertEqual(m, u[2])

        v = [0, 0, -1]
        m = dot(u, v)
        self.assertEqual(m, -3)
        self.assertEqual(m, -u[2])

    def testProjection(self):
        """
        Test the proj function
        """
        console.terse("{0}\n".format(self.testProjection.__doc__))

        from ioflo.aid.vectoring import proj

        u = [1, 2, 3]  # list
        v = [4, 5, 6]
        w = proj(u, v)
        vv = (1.6623376623376622, 2.0779220779220777, 2.493506493506493)
        for e1, e2 in zip(w, vv):
            self.assertAlmostEqual(e1, e2)

        u = [0, 0, 1]
        v = [0, 0, 2]
        w = proj(u, v)
        self.assertEqual(w, (0.0, 0.0, 1.0))

        u = [0, 0, 1]
        v = [0, 0, 0]
        w = proj(u, v)
        self.assertEqual(w, (0, 0, 0))

    def testTrip(self):
        """
        Test the trip function
        """
        console.terse("{0}\n".format(self.testTrip.__doc__))

        from ioflo.aid.vectoring import trip, ccw, cw

        u = (1, 1)
        v = (-1, 1)
        m = trip(u, v)
        self.assertEqual(m, 2)
        l = ccw(u, v)
        self.assertTrue(l)
        r = cw(u, v)
        self.assertFalse(r)

        n = trip(v, u)
        self.assertEqual(n, -m)
        l = ccw(v, u)
        self.assertFalse(l)
        r = cw(v, u)
        self.assertTrue(r)

        p = trip(u, u)
        self.assertEqual(p, 0)
        l = ccw(u, u)
        self.assertFalse(l)
        r = cw(u, u)
        self.assertFalse(r)

    def testTween(self):
        """
        Test the tween functions
        """
        console.terse("{0}\n".format(self.testTween.__doc__))

        from ioflo.aid.vectoring import tween, tween2

        p = (3, 3, 3)
        u = (2, 2, 2)
        v = (4, 4, 4)
        result = tween(p, u, v)
        self.assertTrue(result)

        p = (5, 5, 5)
        u = (2, 2, 2)
        v = (4, 4, 4)
        result = tween(p, u, v)
        self.assertFalse(result)

        p = (1, 1, 1)
        u = (2, 2, 2)
        v = (4, 4, 4)
        result = tween(p, u, v)
        self.assertFalse(result)

        p = (3, 3, 0)
        u = (2, 2, 2)
        v = (4, 4, 4)
        result = tween(p, u, v)
        self.assertFalse(result)

        p = (2, 2, 2)
        u = (2, 2, 2)
        v = (4, 4, 4)
        result = tween(p, u, v)
        self.assertTrue(result)

        p = (4, 4, 4)
        u = (2, 2, 2)
        v = (4, 4, 4)
        result = tween(p, u, v)
        self.assertTrue(result)

        p = (2, 2, 2)
        u = (2, 2, 2)
        v = (2, 2, 2)
        result = tween(p, u, v)
        self.assertTrue(result)

        p = (4, 4, 4)
        u = (2, 2, 2)
        v = (2, 2, 2)
        result = tween(p, u, v)
        self.assertFalse(result)

        p = (3, 3)
        u = (2, 2)
        v = (4, 4)
        result = tween2(p, u, v)
        self.assertTrue(result)

        p = (5, 5)
        u = (2, 2)
        v = (4, 4)
        result = tween2(p, u, v)
        self.assertFalse(result)

        p = (1, 1)
        u = (2, 2)
        v = (4, 4)
        result = tween2(p, u, v)
        self.assertFalse(result)

        p = (3, 0)
        u = (2, 2)
        v = (4, 4)
        result = tween2(p, u, v)
        self.assertFalse(result)

        p = (2, 2)
        u = (2, 2)
        v = (4, 4)
        result = tween2(p, u, v)
        self.assertTrue(result)

        p = (4, 4)
        u = (2, 2)
        v = (4, 4)
        result = tween2(p, u, v)
        self.assertTrue(result)

        p = (2, 2)
        u = (2, 2)
        v = (2, 2)
        result = tween2(p, u, v)
        self.assertTrue(result)

        p = (4, 4)
        u = (2, 2)
        v = (2, 2)
        result = tween2(p, u, v)
        self.assertFalse(result)



    def testCross3Product(self):
        """
        Test the dot function
        """
        console.terse("{0}\n".format(self.testCross3Product.__doc__))

        from ioflo.aid.vectoring import cross3

        u = (1, 2, 3)
        v = (-1, -2, 3)
        w = cross3(u, v)
        self.assertEqual(w, (12, -6, 0))

        u = (1, 1, 0)
        v = (-1, 1, 0)
        w = cross3(u, v)
        self.assertEqual(w, (0, 0, 2))

        w = cross3(v, u)
        self.assertEqual(w, (0, 0, -2))


    def testPointInPolygon(self):
        """
        Test the point in polygon test functions
        """
        console.terse("{0}\n".format(self.testPointInPolygon.__doc__))

        from ioflo.aid.vectoring import wind, inside, insideOnly, outside, outsideOnly, sideOnly

        # counter clockwise wind
        p = (1, 1)
        vs = ((0, 0), (2, 0), (2, 2), (0, 2))  # ccw
        ccw = wind(p, vs)
        self.assertEqual(ccw, 1)
        self.assertTrue(inside(p, vs))
        self.assertTrue(insideOnly(p, vs))
        self.assertFalse(outside(p, vs))
        self.assertFalse(outsideOnly(p, vs))
        self.assertFalse(sideOnly(p, vs))

        # clockwise wind
        vs = ((0, 0), (0, 2), (2, 2), (2, 0)) # cw
        cw = wind(p, vs)
        self.assertEqual(cw, -1)
        self.assertEqual(cw, -ccw)
        self.assertTrue(inside(p, vs))
        self.assertTrue(insideOnly(p, vs))
        self.assertFalse(outside(p, vs))
        self.assertFalse(outsideOnly(p, vs))
        self.assertFalse(sideOnly(p, vs))

        # point not in polygon
        p = (-1, -1)
        vs = ((0, 0), (2, 0), (2, 2), (0, 2))  # ccw
        w = wind(p, vs)
        self.assertEqual(w, 0)
        self.assertFalse(inside(p, vs))
        self.assertFalse(insideOnly(p, vs))
        self.assertTrue(outside(p, vs))
        self.assertTrue(outsideOnly(p, vs))
        self.assertFalse(sideOnly(p, vs))

        vs = ((0, 0), (0, 2), (2, 2), (2, 0)) # cw
        w = wind(p, vs)
        self.assertEqual(w, 0)
        self.assertFalse(inside(p, vs))
        self.assertFalse(insideOnly(p, vs))
        self.assertTrue(outside(p, vs))
        self.assertTrue(outsideOnly(p, vs))
        self.assertFalse(sideOnly(p, vs))

        # point is vertex of polygon
        p = (2, 0)
        vs = ((0, 0), (2, 0), (2, 2), (0, 2))  # ccw
        self.assertTrue(p in vs)
        w = wind(p, vs)
        self.assertEqual(w, 0)
        self.assertTrue(inside(p, vs))
        self.assertFalse(insideOnly(p, vs))
        self.assertTrue(outside(p, vs))
        self.assertFalse(outsideOnly(p, vs))
        self.assertTrue(sideOnly(p, vs))

        vs = ((0, 0), (0, 2), (2, 2), (2, 0))  # cw
        self.assertTrue(p in vs)
        w = wind(p, vs)
        self.assertEqual(w, 0)
        self.assertTrue(inside(p, vs))
        self.assertFalse(insideOnly(p, vs))
        self.assertTrue(outside(p, vs))
        self.assertFalse(outsideOnly(p, vs))
        self.assertTrue(sideOnly(p, vs))

        # point is vertex
        p = (1, 1)
        vs = ((0, 0), (1, 1), (2, 0), (2, 2), (0, 2))  # ccw
        self.assertTrue(p in vs)
        w = wind(p, vs)
        self.assertEqual(w, 0)
        self.assertTrue(inside(p, vs))
        self.assertFalse(insideOnly(p, vs))
        self.assertTrue(outside(p, vs))
        self.assertFalse(outsideOnly(p, vs))
        self.assertTrue(sideOnly(p, vs))

        vs = ((0, 0), (0, 2), (2, 2), (2, 0), (1, 1))  # cw
        self.assertTrue(p in vs)
        w = wind(p, vs)
        self.assertEqual(w, 0)
        self.assertTrue(inside(p, vs))
        self.assertFalse(insideOnly(p, vs))
        self.assertTrue(outside(p, vs))
        self.assertFalse(outsideOnly(p, vs))
        self.assertTrue(sideOnly(p, vs))

        # point on edge side
        p = (1, 0)
        vs = ((0, 0), (2, 0), (2, 2), (0, 2))  # ccw
        w = wind(p, vs)
        self.assertEqual(w, 0)
        self.assertTrue(inside(p, vs))
        self.assertFalse(insideOnly(p, vs))
        self.assertTrue(outside(p, vs))
        self.assertFalse(outsideOnly(p, vs))
        self.assertTrue(sideOnly(p, vs))

        vs = ((0, 0), (0, 2), (2, 2), (2, 0))  # cw
        w = wind(p, vs)
        self.assertEqual(w, 0)
        self.assertTrue(inside(p, vs))
        self.assertFalse(insideOnly(p, vs))
        self.assertTrue(outside(p, vs))
        self.assertFalse(outsideOnly(p, vs))
        self.assertTrue(sideOnly(p, vs))

        # side collinear with p y intercept
        p = (0.5, 1)
        vs = ((0, 0), (1, 1), (2, 1), (2, 2), (0, 2))  # ccw
        w = wind(p, vs)
        self.assertEqual(w, 1)
        self.assertTrue(inside(p, vs))
        self.assertTrue(insideOnly(p, vs))
        self.assertFalse(outside(p, vs))
        self.assertFalse(outsideOnly(p, vs))
        self.assertFalse(sideOnly(p, vs))

        vs = ((0, 0), (0, 2), (2, 2), (2, 1), (1, 1))  # cw
        w = wind(p, vs)
        self.assertEqual(w, -1)
        self.assertTrue(inside(p, vs))
        self.assertTrue(insideOnly(p, vs))
        self.assertFalse(outside(p, vs))
        self.assertFalse(outsideOnly(p, vs))
        self.assertFalse(sideOnly(p, vs))



def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
             'testMagnitude',
             'testAddNegSubMult',
             'testNormalize',
             'testDotProduct',
             'testProjection',
             'testTrip',
             'testTween',
             'testCross3Product',
             'testPointInPolygon',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #console.reinit(verbosity=console.Wordage.concise)

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testBasic')





"""
aiding.py miscellaneous utility functions

"""
from __future__ import absolute_import, division, print_function

import sys
import os
import re

# Import ioflo libs
from .sixing import *
from .consoling import getConsole

console = getConsole()

# For backwards compatibility. In the future import from .byting
from .byting import (binize, unbinize, hexize, unhexize, hexify, unhexify,
                     bytify, unbytify, packify, unpackify, packByte, unpackByte,
                     denary2BinaryStr, dec2BinStr, printHex, printDecimal)

# For backwards compatibility. In the future import from .eventing
from .eventing import tagify, eventify

# For backwards compatibility. In the future import from .checking
from .checking import crc16, crc64

# For backwards compatibility. In the future import from .filing
from .filing import ocfn, load, dump, loadPickle, dumpPickle, loadJson, dumpJson


def repack(n, seq, default=None):
    """ Repacks seq into a generator of len n and returns the generator.
        The purpose is to enable unpacking into n variables.
        The first n-1 elements of seq are returned as the first n-1 elements of the
        generator and any remaining elements are returned in a tuple as the
        last element of the generator
        default (None) is substituted for missing elements when len(seq) < n

        Example:

        x = (1, 2, 3, 4)
        tuple(repack(3, x))
        (1, 2, (3, 4))

        x = (1, 2, 3)
        tuple(repack(3, x))
        (1, 2, (3,))

        x = (1, 2)
        tuple(repack(3, x))
        (1, 2, ())

        x = (1, )
        tuple(repack(3, x))
        (1, None, ())

        x = ()
        tuple(repack(3, x))
        (None, None, ())

    """
    it = iter(seq)
    for _i in range(n - 1):
        yield next(it, default)
    yield tuple(it)

def just(n, seq, default=None):
    """ Returns a generator of just the first n elements of seq and substitutes
        default (None) for any missing elements. This guarantees that a generator of exactly
        n elements is returned. This is to enable unpacking into n varaibles

        Example:

        x = (1, 2, 3, 4)
        tuple(just(3, x))
        (1, 2, 3)
        x = (1, 2, 3)
        tuple(just(3, x))
        (1, 2, 3)
        x = (1, 2)
        tuple(just(3, x))
        (1, 2, None)
        x = (1, )
        tuple(just(3, x))
        (1, None, None)
        x = ()
        tuple(just(3, x))
        (None, None, None)

    """
    it = iter(seq)
    for _i in range(n):
        yield next(it, default)

def reverseCamel(name, lower=True):
    """ Returns camel case reverse of name.
        case change boundaries are the sections which are reversed.
        If lower is True then the initial letter in the reversed name is lower case

        Assumes name is of the correct format to be Python Identifier.
    """
    index = 0
    parts = [[]]
    letters = list(name) # list of the letters in the name
    for c in letters:
        if c.isupper(): #new part
            parts.append([])
            index += 1
        parts[index].append(c.lower())
    parts.reverse()
    parts = ["".join(part) for part in  parts]
    if lower: #camel case with initial lower
        name = "".join(parts[0:1] + [part.capitalize() for part in parts[1:]])
    else: #camel case with initial upper
        name = "".join([part.capitalize() for part in parts])
    return name

def nameToPath(name):
    """ Converts camel case name into full node path where uppercase letters denote
        intermediate nodes in path. Node path ends in dot '.'

        Assumes Name is of the correct format to be Identifier.
    """
    pathParts = []
    nameParts = list(name)
    for c in nameParts:
        if c.isupper():
            pathParts.append('.')
            pathParts.append(c.lower())
        else:
            pathParts.append(c)
    pathParts.append('.')
    path = ''.join(pathParts)
    return path

def isPath(s):
    r"""Returns True if string s is valid Store path name
       Returns False otherwise

       Faster to use precompiled versions in base

       raw string
       this also matches an empty string so need
       r'^([a-zA-Z_][a-zA-Z_0-9]*)?([.][a-zA-Z_][a-zA-Z_0-9]*)*$'

       at least either one of these
       r'^([a-zA-Z_][a-zA-Z_0-9]*)+([.][a-zA-Z_][a-zA-Z_0-9]*)*$'
       r'^([.][a-zA-Z_][a-zA-Z_0-9]*)+$'

       so get
       r'^([a-zA-Z_][a-zA-Z_0-9]*)+([.][a-zA-Z_][a-zA-Z_0-9]*)*$|^([.][a-zA-Z_][a-zA-Z_0-9]*)+$'

       shorthand replace [a-zA-Z_0-9] with  \w which is shorthand for [a-zA-Z_0-9]
       r'^([a-zA-Z_]\w*)+([.][a-zA-Z_]\w*)*$|^([.][a-zA-Z_]\w*)+$'

       ^ anchor to start
       $ anchor to end
       | must either match preceding or succeeding expression
       * repeat previous match zero or more times greedily
       ? repeat previous match zero or one times
       ( ) group
       [ ] char from set of ranges
       [a-zA-Z_] alpha or underscore one and only one
       [a-zA-Z_0-9]* alpha numeric or underscore (zero or more)
       ([a-zA-Z_][a-zA-Z_0-9]*) group made up of one alpha_ and zero or more alphanumeric_
       ([a-zA-Z_][a-zA-Z_0-9]*)? zero or one of the previous group

       ([.][a-zA-Z_][a-zA-Z_0-9]*) group made of one period one alpha_ and zero or more alphanumeric_
       ([.][a-zA-Z_][a-zA-Z_0-9]*)* zero or more of the previous group

       so what it matches.
       if first character is alpha_ then all remaining alphanumeric_ characters will
       match up to but not including first period if any

       from then on it will match groups that start with period one alpha_ and zero
       or more alphanumeric_ until the end

       valid forms
       a
       a1
       .a
       .a1

       a.b
       a1.b2
       .a1.b2
       .a.b

       but not
       .
       a.
       a..b
       ..a
       1.2

    """
    if re.match(r'^([a-zA-Z_]\w*)+([.][a-zA-Z_]\w*)*$|^([.][a-zA-Z_]\w*)+$',s):
        return True
    else:
        return False

def isIdentifier(s):
    r"""Returns True if string s is valid python identifier (variable, attribute etc)
       Returns False otherwise

       how to determine if string is valid python identifier

       r'^[a-zA-Z_]\w*$'
       r'^[a-zA-Z_][a-zA-Z_0-9]*$'  #equivalent \w is shorthand for [a-zA-Z_0-9]

       r' = raw string
       ^ = anchor to start
       [a-zA-Z_] = first char is letter or underscore
       [a-zA-Z_0-9] = next char is letter, underscore, or digit
       * = repeat previous character match greedily
       $ = anchor to end

       How
       import re
       reo = re.compile(r'^[a-zA-Z_]\w*$') #compile is faster
       if reo.match('_hello') is not None: #matched returns match object or None

       #re.match caches compiled pattern string compile so faster after first
       if re.match(r'^[a-zA-Z_]\w*$', '_hello')

       reo = re.compile(r'^[a-zA-Z_][a-zA-Z_0-9]*$')
       reo.match(

    """
    if re.match(r'^[a-zA-Z_]\w*$',s):
        return True
    else:
        return False

def isIdentPub(s):
    """Returns True if string s is valid python public identifier,
       that is, an identifier that does not start with an underscore
       Returns False otherwise
    """
    if re.match(r'^[a-zA-Z]\w*$',s):
        return True
    else:
        return False




"""
byting.py utility functions supporting byte, bit, hex conversion and packing

"""
from __future__ import absolute_import, division, print_function

import sys
import os
import struct
import string

# Import ioflo libs
from .sixing import *
from .odicting import odict

from .consoling import getConsole
console = getConsole()


def binize(n=0, size=8):
    """
    Returns the expanded binary unicode string equivalent of integer n,
    left zero padded using size number of bits

    example binize(11, 4) returns u'1011'
    """
    return "".join([str((n >> y) & 1) for y in range(size-1, -1, -1)])

def unbinize(u=u''):
    """
    Returns integer n equivalent of binary unicode string u

    example: unbinize(u'1011') returns 11
    """
    n = 0
    for bit in u:
        n <<= 1
        n |= 1 if int(bit) else 0
    return n

def hexize(b=b''):
    """
    Returns the expanded hex unicode string equivalent of the bytes string b
    Converts bytes string  b into the unicode hex format equivalent
    Where each byte in bytes b is expanded into the 2 charater hex
    equivalent
    """
    h = u''
    for i in range(len(b)):
        h += "{0:02x}".format(ord(b[i:i+1]))
    return h

def unhexize(h=u''):
    """
    Returns the packed bytes string version of the unicode hex string h
    Converts unicode string h in hex format into  bytes string by compressing
    every two hex characters into 1 byte that is the binary equivalent
    If h does not have an even number of characters then a 0 is first prepended to h

    """
    #remove any non hex characters, any char that is not in '0123456789ABCDEF'
    hh = h  # make copy so iteration not change
    for c in hh:
        if c not in string.hexdigits:
            h = h.replace(c,'') #delete characters

    if len(h) % 2: #odd number of characters
        h = u'0' + h #prepend a zero to make even number

    b = b''
    for i in range(0, len(h), 2):
        s = h[i:i+2]
        b = b + struct.pack('!B', int(s, 16))

    return b

def hexify(b=bytearray([])):
    """
    Returns the unicode string hex equivalent of the bytearray b
    Converts byte in b into into the two character unicode string hex equivalent
    """
    b = bytearray(b)  # just in case bytes
    h = u''
    for byte in b:
        h += "{0:02x}".format(byte)
    return h

def unhexify(h=u''):
    """
    Returns the packed binary bytearray equivalent of the unicode hex string h
    version of the string as bytes
    Where every two hex characters in h are compressed into one byte binary equivalent
    If h does not have an even number of characters then a '0' is first prependedto h
    """
    #remove any non hex characters, any char that is not in '0123456789ABCDEF'
    hh = h #make copy so iteration not change
    for c in hh:
        if c not in string.hexdigits:
            h = h.replace(c,'') #delete characters
    if len(h) % 2: #odd number of characters
        h = u'0' + h #prepend a zero to make even number
    b = bytearray([])
    for i in range(0, len(h), 2):
        s = h[i:i+2]
        b.append(int(s, 16))
    return b

def bytify(n=0, size=1, reverse=False, strict=False):
    """
    Returns bytearray of at least size bytes equivalent of integer n that is
    left zero padded to size bytes. For n positive, if the bytearray
    equivalent of n is longer than size  and strict is False the bytearray
    is extended to the length needed to fully represent n. Otherwise if strict
    is True or n is negative it is truncated to the least significant size bytes.

    Big endian is the default.
    If reverse is true then it reverses the order of the bytes in the resultant
    bytearray for little endian with right zero padding.
    """
    if n < 0 or strict:
        n =  n & (2 ** (size * 8) - 1)
    b = bytearray()
    count = 0
    while n:
        b.insert(0, n & 0xFF)
        count += 1
        n >>=  8
    if (count < size):
        b = bytearray([0]*(size-count)) + b
    if reverse:
        b.reverse()
    return b

def unbytify(b=bytearray([]), reverse=False):
    """
    Returns unsigned integer equivalent of bytearray b
    b may be any iterable of ints including bytes or list of ints.
    Big endian is the default.
    If reverse is true then it reverses the order of the bytes in the byte array
    before conversion for little endian.
    """
    b = bytearray(b)
    if not reverse:  # process MSB first on the right
        b.reverse()
    n = 0
    while b:
        n <<= 8
        n += b.pop()
    return n

def packify(fmt=u'8', fields=[0x00], size=None, reverse=False):
    """
    Packs fields sequence of bit fields into bytearray of size bytes using fmt string.
    Each white space separated field of fmt is the length of the associated bit field
    If not provided size is the least integer number of bytes that hold the fmt.
    If reverse is true reverse the order of the bytes in the byte array before
    returning. This is useful for converting between bigendian and littleendian.

    Assumes unsigned fields values.
    Assumes network big endian so first fields element is high order bits.
    Each field in format string is number of bits for the associated bit field
    Fields with length of 1 are treated as has having boolean truthy field values
       that is,   nonzero is True and packs as a 1
    for 2+ length bit fields the field element is truncated to the number of
       low order bits in the bit field
    if sum of number of bits in fmt less than size bytes then the last byte in
       the bytearray is right zero padded
    if sum of number of bits in fmt greater than size bytes returns exception
    to pad just use 0 value in source field.
    example
    packify("1 3 2 2", (True, 4, 0, 3)). returns bytearry([0xc3])
    """
    tbfl = sum((int(x) for x in fmt.split()))
    if size is None:
        size = (tbfl // 8) + 1 if tbfl % 8 else tbfl // 8

    if not (0 <= tbfl <= (size * 8)):
        raise ValueError("Total bit field lengths in fmt not in [0, {0}]".format(size * 8))

    n = 0
    bfp = 8 * size  # starting bit field position
    bu = 0  # bits used

    for i, bfmt in enumerate(fmt.split()):
        bits = 0x00
        bfl = int(bfmt)
        bu += bfl

        if bfl == 1:
            if fields[i]:
                bits = 0x01
            else:
                bits = 0x00
        else:
            bits = fields[i] & (2**bfl - 1)  # bit-and mask out high order bits

        bits <<= (bfp - bfl) #shift left to bit position less bit field size

        n |= bits  # bit-or in bits
        bfp -= bfl #adjust bit field position for next element

    return bytify(n=n, size=size, reverse=reverse, strict=True)

def packifyInto(b, fmt=u'8', fields=[0x00], size=None, offset=0, reverse=False):
    """
    Packs fields sequence of bit fields using fmt string into bytearray b
    starting at offset and packing into size bytes
    Each white space separated field of fmt is the length of the associated bit field
    If not provided size is the least integer number of bytes that hold the fmt.
    Extends the length of b to accomodate size after offset if not enough.
    Returns actual size of portion packed into.
    The default assumes big endian.
    If reverse is True then reverses the byte order before extending. Useful for
    little endian.

    Assumes unsigned fields values.
    Assumes network big endian so first fields element is high order bits.
    Each field in format string is number of bits for the associated bit field
    Fields with length of 1 are treated as has having boolean truthy field values
       that is,   nonzero is True and packs as a 1
    for 2+ length bit fields the field element is truncated
    to the number of low order bits in the bit field
    if sum of number of bits in fmt less than size bytes then the last byte in
       the bytearray is right zero padded
    if sum of number of bits in fmt greater than size bytes returns exception
    to pad just use 0 value in source field.
    example
    packify("1 3 2 2", (True, 4, 0, 3)). returns bytearry([0xc3])
    """
    tbfl = sum((int(x) for x in fmt.split()))
    if size is None:
        size = (tbfl // 8) + 1 if tbfl % 8 else tbfl // 8

    if not (0 <= tbfl <= (size * 8)):
        raise ValueError("Total bit field lengths in fmt not in [0, {0}]".format(size * 8))

    if len(b) < (offset + size):
        b.extend([0x00]*(offset + size - len(b)))

    n = 0
    bfp = 8 * size  # starting bit field position
    bu = 0  # bits used

    for i, bfmt in enumerate(fmt.split()):
        bits = 0x00
        bfl = int(bfmt)
        bu += bfl

        if bfl == 1:
            if fields[i]:
                bits = 0x01
            else:
                bits = 0x00
        else:
            bits = fields[i] & (2**bfl - 1)  # bit-and mask out high order bits

        bits <<= (bfp - bfl) #shift left to bit position less bit field size

        n |= bits  # bit-or in bits
        bfp -= bfl #adjust bit field position for next element

    bp = bytify(n=n, size=size, reverse=reverse, strict=True)

    b[offset:offset + len(bp)] = bp

    #count = 0
    #while n or (count < size):
        #b[offset + size - 1 - count] = n & 0xFF
        #count += 1
        #n >>=  8

    return size

def unpackify(fmt=u'1 1 1 1 1 1 1 1',
              b=bytearray([0x00]),
              boolean=False,
              size=None,
              reverse=False):
    """
    Returns tuple of unsigned int bit field values that are unpacked from the
    bytearray b according to fmt string. b maybe an integer iterator
    If not provided size is the least integer number of bytes that hold the fmt.
    The default assumes big endian.
    If reverse is True then reverse the byte order of b before unpackifing. This
    is useful for little endian.

    Each white space separated field of fmt is the length of the associated bit field.
    returns unsigned fields values.

    Assumes network big endian so first fmt is high order bits.
    Format string is number of bits per bit field
    If boolean parameter is True then return boolean values for
       bit fields of length 1

    if sum of number of bits in fmt less than 8 * size) then remaining
    bits are returned as additional field in result.

    if sum of number of bits in fmt greater 8 * len(b) returns exception

    example:
    unpackify(u"1 3 2 2", bytearray([0xc3]), False) returns (1, 4, 0, 3)
    unpackify(u"1 3 2 2", 0xc3, True) returns (True, 4, 0, 3)
    """
    b = bytearray(b)
    if reverse:
        b.reverse()

    tbfl = sum((int(x) for x in fmt.split()))
    if size is None:
        size = (tbfl // 8) + 1 if tbfl % 8 else tbfl // 8

    if not (0 <= tbfl <= (size * 8)):
        raise ValueError("Total bit field lengths in fmt not in [0, {0}]".format(size * 8))

    b = b[:size]
    fields = []  # list of bit fields
    bfp = 8 * size  # bit field position
    bu = 0  # bits used
    n = unbytify(b)  # unsigned int equivalent of b

    for i, bfmt in enumerate(fmt.split()):
        bfl = int(bfmt)
        bu += bfl

        mask = (2**bfl - 1) << (bfp - bfl)  # make mask
        bits = n & mask  # mask off other bits
        bits >>= (bfp - bfl)  # right shift to low order bits
        if bfl == 1 and boolean: #convert to boolean
            if bits:
                bits = True
            else:
                bits = False
        fields.append(bits) #assign to fields list
        bfp -= bfl #adjust bit field position for next element

    if bfp != 0:  # remaining bits
        bfl = bfp
        mask = (2**bfl - 1) # make mask
        bits = n & mask  # mask off other bits
        if bfl == 1 and boolean: #convert to boolean
            if bits:
                bits = True
            else:
                bits = False
        fields.append(bits) #assign to fields list
    return tuple(fields) #convert to tuple

def signExtend(x, n=8):
    """
    Returns signed integer that is the sign extention of n bit unsigned integer x
    in twos complement form where x has n significant bits

    This is useful when unpacking bit fields where the bit fields use two's complement
    to represent signed numbers. Assumes the the upper bits (above n) of x are zeros
    Works for both positive and negative x
    """
    m = 1 << (n-1)  # mask
    r = (x ^ m) - m
    return (r)

def packByte(fmt = b'8', fields = [0x00]):
    """Packs fields sequence into one byte using fmt string.

       Each fields element is a bit field and each
       char in fmt is the corresponding bit field length.
       Assumes unsigned fields values.
       Assumes network big endian so first fields element is high order bits.
       Format string is number of bits per bit field
       Fields with length of 1 are treated as has having boolean truthy field values
          that is,   nonzero is True and packs as a 1
       for 2-8 length bit fields the field element is truncated
       to the number of low order bits in the bit field
       if sum of number of bits in fmt less than 8 last bits are padded
       if sum of number of bits in fmt greater than 8 returns exception
       to pad just use 0 value in source field.
       example
       packByte("1322",(True,4,0,3)). returns 0xc3
    """
    fmt = bytes(fmt)
    byte = 0x00
    bfp = 8 #bit field position
    bu = 0 #bits used

    for i in range(len(fmt)):
        bits = 0x00
        bfl = int(fmt[i:i+1])  # slice returns bytes not byte so int works

        if not (0 < bfl <= 8):
            raise ValueError("Bit field length in fmt must be > 0 and <= 8")

        bu += bfl
        if bu > 8:
            raise ValueError("Sum of bit field lengths in fmt must be <= 8")

        if bfl == 1:
            if fields[i]:
                bits = 0x01
            else:
                bits = 0x00
        else:
            bits = fields[i] & (2**bfl - 1) #bit and to mask out high order bits

        bits <<= (bfp - bfl) #shift left to bit position less bit field size

        byte |= bits #or in bits
        bfp -= bfl #adjust bit field position for next element

    console.profuse("Packed byte = {0:#x}\n".format(byte))

    return byte

def unpackByte(fmt = b'11111111', byte = 0x00, boolean = True):
    """unpacks source byte into tuple of bit fields given by fmt string.

       Each char of fmt is a bit field length.
       returns unsigned fields values.
       Assumes network big endian so first fmt is high order bits.
       Format string is number of bits per bit field
       If boolean parameter is True then return boolean values for
          bit fields of length 1

       if sum of number of bits in fmt less than 8 then remaining
       bits returned as additional element in result.

       if sum of number of bits in fmt greater than 8 returns exception
       only low order byte of byte is used.

       example
       unpackByte("1322",0xc3, False ) returns (1,4,0,3)
       unpackByte("1322",0xc3, True ) returns (True,4,0,3)
    """
    fmt = bytes(fmt)
    fields = [] #list of bit fields
    bfp = 8 #bit field position
    bu = 0 #bits used
    byte &= 0xff #get low order byte

    for i in range(len(fmt)):
        bfl = int(fmt[i:i+1])  # slice returns bytes not byte so int works

        if not (0 < bfl <= 8):
            raise ValueError("Bit field length in fmt must be > 0 and <= 8")

        bu += bfl
        if bu > 8:
            raise ValueError("Sum of bit field lengths in fmt must be <= 8")

        mask = (2**bfl - 1) << (bfp - bfl) #make mask
        bits = byte & mask #mask off other bits
        bits >>= (bfp - bfl) #right shift to low order bits
        if bfl == 1 and boolean: #convert to boolean
            if bits:
                bits = True
            else:
                bits = False

        fields.append(bits) #assign to fields list

        bfp -= bfl #adjust bit field position for next element

    return tuple(fields) #convert to tuple

def denary2BinaryStr(n, l=8):
    """
    Convert denary integer n to binary string bs, left pad to length l
    """
    bs = ''
    if n < 0:  raise ValueError("must be a positive integer")
    if n == 0: return '0'
    while n > 0:
        bs = str(n % 2) + bs
        n = n >> 1
    return bs.rjust(l,'0')

def dec2BinStr(n, count=24):
    """ returns the binary formated string of integer n, using count number of digits"""
    return "".join([str((n >> y) & 1) for y in range(count-1, -1, -1)])

def printHex(s, chunk = 0, chunks = 0, silent = False, separator = '.'):
    """prints elements of bytes string s in hex notation.

       chunk is number of bytes per chunk
       0 means no chunking
       chunks is the number of chunks per line
       0 means no new lines

       silent = True means return formatted string but do not print
    """
    if (chunk < 0):
        raise ValueError("invalid size of chunk")

    if (chunks < 0):
        raise ValueError("invalid chunks per line")

    slen = len(s)

    if chunk == 0:
        chunk = slen

    if chunks == 0:
        line = slen
    else:
        line = chunk * chunks

    cc = 0
    ps = ''
    for i in range(len(s)):
        ps += ("%02x" % ord(s[i:i+1]))
        #add space or dot if not end of line or end of string
        if ((i + 1) % line) and ((i+1) % slen):
            if not ((i + 1) % chunk): #end of chunk
                ps += ' ' #space between chunks
            else:
                ps += separator #between bytes in chunk
        elif (i + 1) != slen: # newline if not last line
            ps += '\n' #newline

    if not silent:
        console.terse("{0}\n".format(ps))

    return ps

def printDecimal(s):
    """prints elements of string s in decimal notation.

    """
    ps = ''
    for i in range(len(s)):
        ps = ps + ("%03d." % ord(s[i:i+1]))
    ps = ps[0:-1] #strip trailing .
    print(ps)





# -*- coding: utf-8 -*-
"""
proto base class test package

To run all the unittests:

from ioflo import test
test.run()

"""
import sys
import unittest
import os

from ioflo import test
from ioflo.aid import getConsole

console = getConsole()

console.reinit(verbosity=console.Wordage.concise)

top = os.path.dirname(os.path.abspath(sys.modules.get(__name__).__file__))

if __name__ == "__main__":
    test.run(top)



# -*- coding: utf-8 -*-
"""
ioflo.aio asynchronous io  test package

To run all the unittests:

from ioflo import test
test.run()

"""
import sys
import unittest
import os

from ioflo import test
from ioflo.aid.consoling import getConsole
console = getConsole()
console.reinit(verbosity=console.Wordage.concise)

#top = os.path.dirname(os.path.abspath(sys.modules.get(__name__).__file__))
start = os.path.dirname(os.path.dirname
                      (os.path.abspath
                       (sys.modules.get(__name__).__file__)))

# need top to be above root for relative imports to not go above top level
top = os.path.dirname(os.path.dirname(start))

if __name__ == "__main__":
    test.run(top, start)



"""
meta class and base class utility classes and functions

"""
from __future__ import absolute_import, division, print_function

import sys
from collections.abc import Iterable, Sequence
from abc import ABCMeta
import functools
import inspect
import types

# import ioflo modules
from .sixing import *

from .consoling import getConsole
console = getConsole()


def metaclassify(metaclass):
    """
    Class decorator for creating a class with a metaclass.
    This enables the same syntax to work in both python2 and python3
    python3 does not support
        class name(object):
            __metaclass__ mymetaclass
    python2 does not support
        class name(metaclass=mymetaclass):

    Borrowed from six.py add_metaclass decorator

    Usage:
    @metaclassify(Meta)
    class MyClass(object):
        pass
    That code produces a class equivalent to:

    on Python 3
    class MyClass(object, metaclass=Meta):
        pass

    on Python 2
    class MyClass(object):
        __metaclass__ = MyMeta
    """
    def wrapper(cls):
        originals = cls.__dict__.copy()
        originals.pop('__dict__', None)
        originals.pop('__weakref__', None)
        for slots_var in originals.get('__slots__', ()):
            originals.pop(slots_var)
        return metaclass(cls.__name__, cls.__bases__, originals)
    return wrapper

@metaclassify(ABCMeta)
class NonStringIterable:
    """ Allows isinstance check for iterable that is not a string
    """
    #__metaclass__ = ABCMeta

    @classmethod
    def __subclasshook__(cls, C):
        if cls is NonStringIterable:
            if (not issubclass(C, (str, bytes)) and issubclass(C, Iterable)):
                return True
        return NotImplemented

@metaclassify(ABCMeta)
class NonStringSequence:
    """ Allows isinstance check for sequence that is not a string
    """
    #__metaclass__ = ABCMeta

    @classmethod
    def __subclasshook__(cls, C):
        if cls is NonStringSequence:
            if (not issubclass(C, (str, bytes)) and issubclass(C, Sequence)):
                return True
        return NotImplemented

def nonStringIterable(obj):
    """
    Returns True if obj is non-string iterable, False otherwise

    Future proof way that is compatible with both Python3 and Python2 to check
    for non string iterables.

    Faster way that is less future proof
    return (hasattr(x, '__iter__') and not isinstance(x, (str, bytes)))
    """
    return (not isinstance(obj, (str, bytes)) and isinstance(obj, Iterable))

def nonStringSequence(obj):
    """
    Returns True if obj is non-string sequence, False otherwise

    Future proof way that is compatible with both Python3 and Python2 to check
    for non string sequences.

    """
    return (not isinstance(obj, (str, bytes)) and isinstance(obj, Sequence) )


def isIterator(obj):
    """
    Returns True if obj is an iterator object, that is,

    has an __iter__ method
    has a __next__ method
    .__iter__ is callable and returns obj

    Otherwise returns False

    """
    if (hasattr(obj, "__iter__") and
        hasattr(obj, "__next__") and
        callable(obj.__iter__) and
        obj.__iter__() is obj
       ):
        return True
    return False



from collections.abc import Generator

def attributize(genfunc):
    """
    Python generators do not support adding attributes.
    Adding support for attributes provides a way to pass information
    from a WSGI App that returns a generator to a WSGI server via the generator

    This decorator takes a Duck Typing approach to decorating
    a generator function or method that returns a new function type instance that
    when called will return a generator like object that supports attributes.
    the new object wrapper or skin acts like a generator but with attributes.

    If genfunc is a generator function then a reference to this wrapper/skin
         is injected as the first positional argument to the orginal
         generator function.
    If genfunc is a generator method, that is, its first parameter is 'self'
        then a reference to this wrapper/skin is injected as the second
        positional argument to the original generator method

    Usage:
    # generator function
    @classing.attributize
    def bar(skin, req=None, rep=None):
        skin._status = 400
        skin._headers = odict(example="Hi")
        yield b""
        yield b""
        yield b"Hello There"
        return b"Goodbye"

    gen = bar()
    msg = next(gen)  # attributes set after first next
    gen._status
    gen._headers

    # generator method
    class R:
        @classing.attributize
        def bar(self, skin, req=None, rep=None):
            self.name = "Peter"
            skin._status = 400
            skin._headers = odict(example="Hi")
            yield b""
            yield b""
            yield b"Hello There " + self.name.encode()
            return b"Goodbye"

    r = R()
    gen = r.bar()
    msg = next(gen)   # attributes set after first next
    gen._status
    gen._headers

    Adding attributes to this injected reference makes them available
    as attributes of the resultant skin

    The new type is AttributiveGenerator

    Parameters:
        genfunc is either
            a generator function that returns a generator object
            a generator method that return a generator object

    Unlike Python functions, Python generators do not support attributes and the
    generator locals dict at .gi_frame.f_locals dissappears once the generator
    is complete so its inconvenient.

    Attributes of generator objects.
    ['.__next__', '__iter__', 'close', 'gi_code', 'gi_frame', 'gi_running',
    'gi_yieldfrom', 'send', 'throw']
    """

    def wrapper(*args, **kwargs):
        """
        When called returns instance of AttributiveGenerator instead of generator.
        """
        def __iter__(self):
            return self

        def send(self):
            raise NotImplementedError

        def throw(self):
            raise NotImplementedError

        tdict = { '__iter__': __iter__, 'send': send, 'throw':  throw,}
        # use type to create dynamic instance of class AttributiveGenerator
        #spec = {'__iter__': lambda self: self, 'send': ,}
        AG = type("AttributiveGenerator", (Generator,), tdict)
        ag = AG()  # create  instance so we can inject it into genfunc

        fargs = inspect.getfullargspec(genfunc).args
        if fargs and fargs[0] == 'self':
            gen = genfunc(args[0], ag, *args[1:], **kwargs)
        else:
            gen = genfunc(ag, *args, **kwargs)  # create generator insert ag ref

        # now add to class references to gen attributes "duckify"
        for attr in ('__next__', 'close', 'send', 'throw',
                     'gi_code', 'gi_frame', 'gi_running', 'gi_yieldfrom'):
            setattr(AG, attr, getattr(gen, attr))

        functools.update_wrapper(wrapper=ag, wrapped=gen)
        return ag
    return wrapper



# -*- coding: utf-8 -*-
"""
Unittests
"""
from __future__ import absolute_import, division, print_function

import sys

import unittest

import os

from binascii import hexlify

# Import ioflo libs
from ioflo.aid.sixing import *
from ioflo.aid.byting import hexify, bytify, unbytify, packify, unpackify
from ioflo.aid import getConsole

console = getConsole()

from ioflo.aio.proto import exchanging, devicing, stacking

def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    console.reinit(verbosity=console.Wordage.concise)

class BasicTestCase(unittest.TestCase):
    """
    Test Case
    """

    def setUp(self):
        """

        """
        console.reinit(verbosity=console.Wordage.profuse)

    def tearDown(self):
        """

        """
        console.reinit(verbosity=console.Wordage.concise)

    def testExchange(self):
        """
        Test Exchange class
        """
        console.terse("{0}\n".format(self.testExchange.__doc__))

        stack = stacking.Stack()
        exchange = exchanging.Exchange(stack=stack)
        self.assertIs(exchange.stack, stack)
        self.assertEqual(len(exchange.uid), 24)
        self.assertEqual(exchange.name, 'Exchange')
        self.assertIs(exchange.device, None)


        exchanger = exchanging.Exchanger(stack=stack)
        self.assertIs(exchanger.stack, stack)
        self.assertEqual(len(exchanger.uid), 24)
        self.assertEqual(exchanger.name, 'Exchanger')
        self.assertIs(exchanger.device, None)

        exchangent = exchanging.Exchangent(stack=stack)
        self.assertIs(exchangent.stack, stack)
        self.assertEqual(len(exchanger.uid), 24)
        self.assertEqual(exchangent.name, 'Exchangent')
        self.assertIs(exchangent.device, None)


def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
             'testExchange',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testPart')






"""
uxd async io (nonblocking) module

"""
from __future__ import absolute_import, division, print_function

import sys
import os
import socket
import errno
from binascii import hexlify

# Import ioflo libs
from ...aid.sixing import *
from ...aid.consoling import getConsole

console = getConsole()

class SocketUxdNb(object):
    """
    Class to manage non blocking io on UXD (unix domain) socket.
    Use instance method .close() to close socket
    """

    def __init__(self, ha=None, umask=None, bufsize = 1024, wlog=None):
        """
        Initialization method for instance.

        ha = uxd file name
        umask = umask for uxd file
        bufsize = buffer size
        """
        self.ha = ha  # uxd host address string name
        self.umask = umask
        self.bs = bufsize
        self.wlog = wlog

        self.ss = None  # server's socket needs to be opened
        self.opened = False

    def actualBufSizes(self):
        """
        Returns duple of the the actual socket send and receive buffer size
        (send, receive)
        """
        if not self.ss:
            return (0, 0)

        return (self.ss.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF),
                self.ss.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF))

    def open(self):
        """
        Opens socket in non blocking mode.

        if socket not closed properly, binding socket gets error
            socket.error: (48, 'Address already in use')
        """
        #create socket ss = server socket
        self.ss = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)

        # make socket address reusable.
        # the SO_REUSEADDR flag tells the kernel to reuse a local socket in
        # TIME_WAIT state, without waiting for its natural timeout to expire.
        self.ss.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if self.ss.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF) <  self.bs:
            self.ss.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, self.bs)
        if self.ss.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF) < self.bs:
            self.ss.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, self.bs)
        self.ss.setblocking(0) #non blocking socket

        oldumask = None
        if self.umask is not None: # change umask for the uxd file
            oldumask = os.umask(self.umask) # set new and return old

        #bind to Host Address Port
        try:
            self.ss.bind(self.ha)
        except socket.error as ex:
            if not ex.errno == errno.ENOENT: # No such file or directory
                console.terse("socket.error = {0}\n".format(ex))
                return False
            try:
                os.makedirs(os.path.dirname(self.ha))
            except OSError as ex:
                console.terse("OSError = {0}\n".format(ex))
                return False
            try:
                self.ss.bind(self.ha)
            except socket.error as ex:
                console.terse("socket.error = {0}\n".format(ex))
                return False

        if oldumask is not None: # restore old umask
            os.umask(oldumask)

        self.ha = self.ss.getsockname() #get resolved ha after bind
        self.opened = True
        return True

    def reopen(self):
        """
        Idempotently open socket by closing first if need be
        """
        self.close()
        return self.open()

    def close(self):
        """
        Closes  socket.
        """
        if self.ss:
            self.ss.close() #close socket
            self.ss = None
            self.opened = False

        try:
            os.unlink(self.ha)
        except OSError:
            if os.path.exists(self.ha):
                raise

    def receive(self):
        """
        Perform non blocking receive on  socket.
        Returns tuple of form (data, sa)
        If no data then returns ('',None)
        """
        try:
            #sa = source address tuple (sourcehost, sourceport)
            data, sa = self.ss.recvfrom(self.bs)
        except socket.error as ex:
            # ex.args[0] is always ex.errno for better compat
            if ex.args[0]  in (errno.EAGAIN, errno.EWOULDBLOCK):
                return (b'', None) #receive has nothing empty string for data
            else:
                emsg = "socket.error = {0}: receiving at {1}\n".format(ex, self.ha)
                console.profuse(emsg)
                raise #re raise exception ex1

        if console._verbosity >= console.Wordage.profuse:  # faster to check
            try:
                load = data.decode("UTF-8")
            except UnicodeDecodeError as ex:
                load = "0x{0}".format(hexlify(data).decode("ASCII"))
            cmsg = ("Server at {0}, received from {1}:\n------------\n"
                    "{2}\n\n".format(self.ha, sa, load))
            console.profuse(cmsg)

        if self.wlog:
            self.wlog.writeRx(sa, data)

        return (data, sa)

    def send(self, data, da):
        """Perform non blocking send on  socket.

           data is string in python2 and bytes in python3
           da is destination address tuple (destHost, destPort)
        """
        try:
            result = self.ss.sendto(data, da) #result is number of bytes sent
        except socket.error as ex:
            emsg = "socket.error = {0}: sending from {1} to {2}\n".format(ex, self.ha, da)
            console.profuse(emsg)
            result = 0
            raise

        if console._verbosity >=  console.Wordage.profuse:
            try:
                load = data[:result].decode("UTF-8")
            except UnicodeDecodeError as ex:
                load = "0x{0}".format(hexlify(data[:result]).decode("ASCII"))
            cmsg = ("Server at {0}, sent {1} bytes to {2}:\n------------\n"
                    "{3}\n\n".format(self.ha, result, da, load))
            console.profuse(cmsg)

        if self.wlog:
            self.wlog.writeTx(da, data)

        return result


PeerUxd = SocketUxdNb  # alias



"""
Asynchronous (nonblocking) serial io

"""
from __future__ import absolute_import, division, print_function

import sys
import os
import errno
from collections import deque

# Import ioflo libs
from ...aid.sixing import *
from ...aid.consoling import getConsole

console = getConsole()

class ConsoleNb(object):
    """
    Class to manage non blocking io on serial console.

    Opens non blocking read file descriptor on console
    Use instance method close to close file descriptor
    Use instance methods getline & put to read & write to console
    Needs os module
    """

    def __init__(self):
        """Initialization method for instance.

        """
        self.fd = None #console file descriptor needs to be opened

    def open(self, port='', canonical=True):
        """
        Opens fd on terminal console in non blocking mode.

        port is the serial port device path name
        or if '' then use os.ctermid() which
        returns path name of console usually '/dev/tty'

        canonical sets the mode for the port. Canonical means no characters
        available until a newline

        os.O_NONBLOCK makes non blocking io
        os.O_RDWR allows both read and write.
        os.O_NOCTTY don't make this the controlling terminal of the process
        O_NOCTTY is only for cross platform portability BSD never makes it the
        controlling terminal

        Don't use print at same time since it will mess up non blocking reads.

        Default is canonical mode so no characters available until newline
        need to add code to enable  non canonical mode

        It appears that canonical mode only applies to the console. For other
        serial ports the characters are available immediately
        """
        if not port:
            port = os.ctermid() #default to console

        try:
            self.fd = os.open(port, os.O_NONBLOCK | os.O_RDWR | os.O_NOCTTY)
        except OSError as ex:
            console.terse("os.error = {0}\n".format(ex))
            return False
        return True

    def close(self):
        """Closes fd.

        """
        if self.fd:
            os.close(self.fd)
            self.fd = None

    def getLine(self,bs = 80):
        """Gets nonblocking line from console up to bs characters including newline.

           Returns empty string if no characters available else returns line.
           In canonical mode no chars available until newline is entered.
        """
        line = ''
        try:
            line = os.read(self.fd, bs)
        except OSError as ex1:  #if no chars available generates exception
            try: #need to catch correct exception
                errno = ex1.args[0] #if args not sequence get TypeError
                if errno == 35:
                    pass #No characters available
                else:
                    raise #re raise exception ex1
            except TypeError as ex2:  #catch args[0] mismatch above
                raise ex1 #ignore TypeError, re-raise exception ex1

        return line

    def put(self, data = '\n'):
        """Writes data string to console.

        """
        return(os.write(self.fd, data))

class DeviceNb(object):
    """
    Class to manage non blocking IO on serial device port.

    Opens non blocking read file descriptor on serial port
    Use instance method close to close file descriptor
    Use instance methods get & put to read & write to serial device
    Needs os module
    """

    def __init__(self, port=None, speed=9600, bs=1024):
        """
        Initialization method for instance.

        port = serial device port path string
        speed = serial port speed in bps
        bs = buffer size for reads
        """
        self.fd = None #serial device port file descriptor, must be opened first
        self.port = port or os.ctermid() #default to console
        self.speed = speed or 9600
        self.bs = bs or 1024
        self.opened = False

    def open(self, port=None, speed=None, bs=None):
        """
        Opens fd on serial port in non blocking mode.

        port is the serial port device path name or
        if '' then use os.ctermid() which
        returns path name of console usually '/dev/tty'

        os.O_NONBLOCK makes non blocking io
        os.O_RDWR allows both read and write.
        os.O_NOCTTY don't make this the controlling terminal of the process
        O_NOCTTY is only for cross platform portability BSD never makes it the
        controlling terminal

        Don't use print and console at same time since it will mess up non blocking reads.

        Raw mode

        def setraw(fd, when=TCSAFLUSH):
        Put terminal into a raw mode.
        mode = tcgetattr(fd)
        mode[IFLAG] = mode[IFLAG] & ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON)
        mode[OFLAG] = mode[OFLAG] & ~(OPOST)
        mode[CFLAG] = mode[CFLAG] & ~(CSIZE | PARENB)
        mode[CFLAG] = mode[CFLAG] | CS8
        mode[LFLAG] = mode[LFLAG] & ~(ECHO | ICANON | IEXTEN | ISIG)
        mode[CC][VMIN] = 1
        mode[CC][VTIME] = 0
        tcsetattr(fd, when, mode)


        # set up raw mode / no echo / binary
        cflag |=  (TERMIOS.CLOCAL|TERMIOS.CREAD)
        lflag &= ~(TERMIOS.ICANON|TERMIOS.ECHO|TERMIOS.ECHOE|TERMIOS.ECHOK|TERMIOS.ECHONL|
                     TERMIOS.ISIG|TERMIOS.IEXTEN) #|TERMIOS.ECHOPRT
        for flag in ('ECHOCTL', 'ECHOKE'): # netbsd workaround for Erk
            if hasattr(TERMIOS, flag):
                lflag &= ~getattr(TERMIOS, flag)

        oflag &= ~(TERMIOS.OPOST)
        iflag &= ~(TERMIOS.INLCR|TERMIOS.IGNCR|TERMIOS.ICRNL|TERMIOS.IGNBRK)
        if hasattr(TERMIOS, 'IUCLC'):
            iflag &= ~TERMIOS.IUCLC
        if hasattr(TERMIOS, 'PARMRK'):
            iflag &= ~TERMIOS.PARMRK

        """
        if port is not None:
            self.port = port
        if speed is not None:
            self.speed = speed
        if bs is not None:
            self.bs = bs

        self.fd = os.open(self.port, os.O_NONBLOCK | os.O_RDWR | os.O_NOCTTY)

        system = platform.system()

        if (system == 'Darwin') or (system == 'Linux'): #use termios to set values
            import termios

            iflag, oflag, cflag, lflag, ispeed, ospeed, cc = range(7)

            settings = termios.tcgetattr(self.fd)
            #print(settings)

            settings[lflag] = (settings[lflag] & ~termios.ICANON)

            settings[lflag] = (settings[lflag] & ~termios.ECHO) # no echo

            #ignore carriage returns on input
            #settings[iflag] = (settings[iflag] | (termios.IGNCR)) #ignore cr

            # 8N1 8bit word no parity one stop bit nohardware handshake ctsrts
            # to set size have to mask out(clear) CSIZE bits and or in size
            settings[cflag] = ((settings[cflag] & ~termios.CSIZE) | termios.CS8)
            # no parity clear PARENB
            settings[cflag] = (settings[cflag] & ~termios.PARENB)
            #one stop bit clear CSTOPB
            settings[cflag] = (settings[cflag] & ~termios.CSTOPB)
            #no hardware handshake clear crtscts
            settings[cflag] = (settings[cflag] & ~termios.CRTSCTS)

            # in linux the speed flag does not equal value so always set it
            speedattr = "B{0}".format(self.speed)  # convert numeric speed to attribute name string
            speed = getattr(termios, speedattr)
            settings[ispeed] = speed
            settings[ospeed] = speed

            termios.tcsetattr(self.fd, termios.TCSANOW, settings)
            #print(settings)

        self.opened = True

    def reopen(self):
        """
        Idempotently open serial device port
        """
        self.close()
        return self.open()

    def close(self):
        """Closes fd.

        """
        if self.fd:
            os.close(self.fd)
            self.fd = None
            self.opened = False

    def receive(self):
        """
        Reads nonblocking characters from serial device up to bs characters
        Returns empty bytes if no characters available else returns all available.
        In canonical mode no chars are available until newline is entered.
        """
        data = b''
        try:
            data = os.read(self.fd, self.bs)  #if no chars available generates exception
        except OSError as ex1:  # ex1 is the target instance of the exception
            if ex1.errno == errno.EAGAIN: #BSD 35, Linux 11
                pass #No characters available
            else:
                raise #re raise exception ex1

        return data

    def send(self, data=b'\n'):
        """
        Writes data bytes to serial device port.
        Returns number of bytes sent
        """
        try:
            count = os.write(self.fd, data)
        except OSError as ex1:  # ex1 is the target instance of the exception
            if ex1.errno == errno.EAGAIN: #BSD 35, Linux 11
                count = 0  # buffer full can't write
            else:
                raise #re raise exception ex1

        return count


class SerialNb(object):
    """
    Class to manage non blocking IO on serial device port using pyserial

    Opens non blocking read file descriptor on serial port
    Use instance method close to close file descriptor
    Use instance methods get & put to read & write to serial device
    Needs os module
    """

    def __init__(self, port=None, speed=9600, bs=1024):
        """
        Initialization method for instance.

        port = serial device port path string
        speed = serial port speed in bps
        bs = buffer size for reads


        """
        self.serial = None  # Serial instance
        self.port = port or os.ctermid() #default to console
        self.speed = speed or 9600
        self.bs = bs or 1024
        self.opened = False

    def open(self, port=None, speed=None, bs=None):
        """
        Opens fd on serial port in non blocking mode.

        port is the serial port device path name or
        if None then use os.ctermid() which returns path name of console
        usually '/dev/tty'
        """
        if port is not None:
            self.port = port
        if speed is not None:
            self.speed = speed
        if bs is not None:
            self.bs = bs

        import serial  # import pyserial
        self.serial = serial.Serial(port=self.port,
                                    baudrate=self.speed,
                                    timeout=0,
                                    writeTimeout=0)
        #self.serial.nonblocking()
        self.serial.reset_input_buffer()
        self.opened = True

    def reopen(self):
        """
        Idempotently open serial device port
        """
        self.close()
        return self.open()

    def close(self):
        """
        Closes .serial
        """
        if self.serial:
            self.serial.reset_output_buffer()
            self.serial.close()
            self.serial = None
            self.opened = False

    def receive(self):
        """
        Reads nonblocking characters from serial device up to bs characters
        Returns empty bytes if no characters available else returns all available.
        In canonical mode no chars are available until newline is entered.
        """
        data = b''
        try:
            data = self.serial.read(self.bs)  #if no chars available generates exception
        except OSError as ex1:  # ex1 is the target instance of the exception
            if ex1.errno == errno.EAGAIN: #BSD 35, Linux 11
                pass #No characters available
            else:
                raise #re raise exception ex1

        return data

    def send(self, data=b'\n'):
        """
        Writes data bytes to serial device port.
        Returns number of bytes sent
        """
        try:
            count = self.serial.write(data)
        except OSError as ex1:  # ex1 is the target instance of the exception
            if ex1.errno == errno.EAGAIN: #BSD 35, Linux 11
                count = 0  # buffer full can't write
            else:
                raise #re raise exception ex1

        return count

class Driver(object):
    """
    Nonblocking Serial Device Port Driver
    """

    def __init__(self,
                 name=u'',
                 uid=0,
                 port=None,
                 speed=9600,
                 bs=1024,
                 server=None):
        """
        Initialization method for instance.

        Parameters:
            name = user friendly name for driver
            uid = unique identifier for driver
            port = serial device port path string
            speed = serial port speed in bps
            canonical = canonical mode True or False
            bs = buffer size for reads
            server = serial port device server if any

        Attributes:
           name = user friendly name for driver
           uid = unique identifier for driver
           server = serial device server nonblocking
           txes = deque of data bytes to send
           rxbs = bytearray of data bytes received

        """
        self.name = name
        self.uid = uid

        if not server:
            try:
                import serial
                self.server = SerialNb(port=port,
                                       speed=speed,
                                       bs=bs)

            except ImportError as  ex:
                console.terse("Error: importing pyserial\n{0}\n".format(ex))
                self.server = DeviceNb(port=port,
                                       speed=speed,
                                       bs=bs)
        else:
            self.server = server

        self.txes = deque()  # deque of data to send
        self.rxbs = bytearray()  # byte array of data received

    def serviceReceives(self):
        """
        Service receives until no more
        """
        while self.server.opened:
            data = self.server.receive()  # bytes
            if not data:
                break
            self.rxbs.extend(data)

    def serviceReceiveOnce(self):
        '''
        Retrieve from server only one reception
        '''
        if self.server.opened:
            data = self.server.receive()
            if data:
                self.rxbs.extend(data)

    def clearRxbs(self):
        """
        Clear .rxbs
        """
        del self.rxbs[:]

    def scan(self, start):
        """
        Returns offset of given start byte in self.rxbs
        Returns None if start is not given or not found
        If strip then remove any bytes before offset
        """
        offset = self.rxbs.find(start)
        if offset < 0:
            return None
        return offset

    def tx(self, data):
        '''
        Queue data onto .txes
        '''
        self.txes.append(data)

    def _serviceOneTx(self):
        """
        Handle one tx data
        """
        data = self.txes.popleft()
        count = self.server.send(data)
        if count < len(data):  # put back unsent portion
            self.txes.appendleft(data[count:])
            return False  # blocked
        console.profuse("{0}: Sent: {1}\n".format(self.name, data))
        return True  # send more

    def serviceTxes(self):
        """
        Service txes data
        """
        while self.txes and self.server.opened:
            again = self._serviceOneTx()
            if not again:
                break  # blocked try again later

    def serviceTxOnce(self):
        '''
        Service one data on the .txes deque to send through device
        '''
        if self.txes and self.server.opened:
            self._serviceOneTx()




# -*- coding: utf-8 -*-
"""
Unit Test Template
"""
from __future__ import absolute_import, division, print_function

import sys

import unittest

import os
import array

from ioflo.aid.sixing import *
from ioflo.test import testing
from ioflo.aid.consoling import getConsole


console = getConsole()

from ioflo.aid import quaternioning


def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    pass


class BasicTestCase(unittest.TestCase):
    """
    Example TestCase
    """

    def setUp(self):
        """
        Call super if override so House Framer and Frame are setup correctly
        """
        super(BasicTestCase, self).setUp()

    def tearDown(self):
        """
        Call super if override so House Framer and Frame are torn down correctly
        """
        super(BasicTestCase, self).tearDown()

    def testQMag(self):
        """
        Test the qmag and qmag2 functions
        """
        console.terse("{0}\n".format(self.testQMag.__doc__))

        from ioflo.aid.quaternioning import qmag, qmag2

        q = [1, 1, 2, 3]  # list
        m = qmag(q)
        self.assertAlmostEqual(m, 3.872983346207417)

        m2 = qmag2(q)
        self.assertEqual(m2, 15)

        q = (1, 1, 2, 3)  # tuple
        m = qmag(q)
        self.assertAlmostEqual(m, 3.872983346207417)

        m2 = qmag2(q)
        self.assertEqual(m2, 15)

        q = [0, 0, 0, 0]
        m = qmag(q)
        self.assertEqual(m, 0.0)

        m2 = qmag2(q)
        self.assertEqual(m2, 0.0)

        q = [1, 0, 0, 0]
        m = qmag(q)
        self.assertEqual(m, 1.0)

        m2 = qmag2(q)
        self.assertEqual(m2, 1)

    def testQNorm(self):
        """
        Test the qnorm function
        """
        console.terse("{0}\n".format(self.testQNorm.__doc__))

        from ioflo.aid.quaternioning import qnorm, qmag

        q = [1, 1, 2, 3]  # list
        n = qnorm(q)
        m = qmag(n)
        self.assertAlmostEqual(m, 1.0)

        z = (0.2581988897471611, 0.2581988897471611, 0.5163977794943222, 0.7745966692414834)
        for e1, e2 in zip(n, z):
            self.assertAlmostEqual(e1, e2)

        q = [0, 0, 0, 0]  # special case
        n = qnorm(q)
        m = qmag(n)
        self.assertAlmostEqual(m, 0.0)

        q = [1, 0, 0, 0]
        n = qnorm(q)
        m = qmag(n)
        self.assertEqual(m, 1.0)

        z = [1.0, 0.0, 0.0, 0.0]
        for e1, e2 in zip(n, z):
            self.assertEqual(e1, e2)

        q = (0.0, 0.0, 0.49999999999999994, 0.8660254037844386)
        n = qnorm(q, check=True)
        m = qmag(n)
        self.assertEqual(m, 1.0)

    def testQConj(self):
        """
        Test the qconj function
        """
        console.terse("{0}\n".format(self.testQConj.__doc__))

        from ioflo.aid.quaternioning import qnorm, qmag, qconj

        q = [1, 1, 2, 3]  # list
        cq = qconj(q)

        z = (1, -1, -2, -3)
        for e1, e2 in zip(cq, z):
            self.assertEqual(e1, e2)

        q = [0, 0, 0, 0]  # special case
        cq = qconj(q)
        for e1, e2 in zip(cq, q):
            self.assertEqual(e1, e2)


        q = [1, 0, 0, 0] # special case
        cq = qconj(q)
        for e1, e2 in zip(cq, q):
            self.assertEqual(e1, e2)


    def testQMul(self):
        """
        Test the qmul function
        """
        console.terse("{0}\n".format(self.testQMul.__doc__))

        from ioflo.aid.quaternioning import qnorm, qmag, qconj, qmul

        q1 = (1, 2, 3, 4)
        q2 = (4, 3, 2, 1)
        q = qmul(q1, q2)

        qq = (-12, 6, 24, 12)
        for e1, e2 in zip(q, qq):
            self.assertEqual(e1, e2)

        rq = qmul(q2, q1)  # not commutative

        qq = (-12, 16, 4, 22)
        for e1, e2 in zip(rq, qq):
            self.assertEqual(e1, e2)

        q1 = (0, 0, 0, 0)  # zero case
        q2 = (4, 3, 2, 1)
        q = qmul(q1, q2)

        qq = (0, 0, 0, 0)
        for e1, e2 in zip(q, qq):
            self.assertEqual(e1, e2)

        q1 = (1, 0, 0, 0)  # identity case
        q2 = (5, 4 , 3, 2)
        q = qmul(q1, q2)

        for e1, e2 in zip(q, q2):
            self.assertEqual(e1, e2)

        q1 = (1, 0, 0, 0)  # identity case commutes
        q2 = (5, 4 , 3, 2)
        q = qmul(q2, q1)

        for e1, e2 in zip(q, q2):
            self.assertEqual(e1, e2)

        q1 = (5, 4, 3, 2)
        q2 = qconj(q1)
        q = qmul(q1, q2)  # qq*

        qq = (54, 0, 0, 0)
        for e1, e2 in zip(q, qq):
            self.assertEqual(e1, e2)

        q3 = qmul(q2, q1)  # q*q == qq*

        for e1, e2 in zip(q, q3):
            self.assertEqual(e1, e2)

        q1 = qnorm((5, 4, 3, 2))  # unit vector
        q2 = qconj(q1)  # conjugate of unit vector is inverse
        q = qmul(q1, q2)

        qq = (1.0, 0, 0, 0)
        for e1, e2 in zip(q, qq):
            self.assertEqual(e1, e2)



def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
             'testQMag',
             'testQNorm',
             'testQConj',
             'testQMul',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #console.reinit(verbosity=console.Wordage.concise)

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testBasic')





"""aiding.py constants and basic functions

"""
from __future__ import absolute_import, division, print_function

import math

# Import ioflo libs
from .sixing import *
from ..base import excepting

from .consoling import getConsole
console = getConsole()


def blend0(d=0.0, u=1.0, s=1.0):
    """
       blending function trapezoid
       d = delta x = xabs - xdr
       u = uncertainty radius of xabs estimate error
       s = tuning scale factor

       returns blend
    """
    d = float(abs(d))
    u = float(abs(u))
    s = float(abs(s))
    v = d - u #offset by radius

    if v >= s:  #first so if s == 0 catches here so no divide by zero below
        b = 0.0
    elif v <= 0.0:
        b = 1.0
    else: # 0 < v < s
        b = 1.0 - (v / s)

    return b

Blend0 = blend0


def blend1(d=0.0, u=1.0, s=1.0):
    """
       blending function pisig
       d = delta x = xabs - xdr
       u = uncertainty radius of xabs estimate error
       s = tuning scale factor

       returns blend
    """
    v = float(abs(u * s)) #scale uncertainty radius make sure positive
    a = float(abs(d)) #symmetric about origin

    if a >= v or v == 0.0 : #outside uncertainty radius accept delta
        b = 1.0
    elif a < v/2.0: # inside 1/2 uncertainty radius closer to 0
        b = 2.0 * (a * a)/(v * v)
    else: #greater than 1/2 uncertainty radius closer to 1
        b = 1.0 - (2.0 * (a - v) * (a - v))/ (v * v)

    return b

Blend1 = blend1


def blend2(d=0.0, u=1.0, s=5.0):
    """
       blending function gaussian
       d = delta x = xabs - xdr
       u = uncertainty radius of xabs estimate error
       s = tuning scale factor

       returns blend
    """
    d = float(d)
    u = float(u)
    s = float(abs(s)) # make sure positive

    b = 1.0 - math.exp( - s * (d * d)/(u * u))

    return b

Blend2 = blend2


def blend3(d=0.0, u=1.0, s=0.05):
    """
       blending function polynomial
       d = delta x = xabs - xdr
       u = uncertainty radius of xabs estimate error
       s = tuning scale factor

       returns blend
    """
    d = float(d)
    u = float(u)
    s = min(1.0,float(abs(s))) # make sure positive <= 1.0

    b = 1.0 - s ** ((d * d)/(u * u))

    return b

Blend3 = blend3



def blendConcaveInc(d=0.0, u=1.0, s=1.0, h=1.0):
    """
       blending function increasing concave
       d = delta x = xabs - xdr
       u = uncertainty radius of xabs estimate error
       s = tuning scale factor

       eq 3.12

       returns blend
    """
    d = float(d)
    u = float(u)
    s = float(s)
    
    m = (2 * s - u - d) 
    if m == 0:
        return h
    
    if d >= s:
        return h
    else:
        b = h * (s - u) / m    
    
    return b

BlendConcaveInc = blendConcaveInc


def blendConcaveDec(d=0.0, u=2.0, s=1.0, h=1.0):
    """
       blending function decreasing concave
       d = delta x = xabs - xdr
       u = uncertainty radius of xabs estimate error
       s = tuning scale factor

       eq 3.13

       returns blend
    """
    d = float(d) 
    u = float(u)
    s = float(s)
    
    m = (u - 2 * s + d) 
    if m == 0:
        return h
    
    if d <= s:
        return h
    else:
        b = (h * (u - s)) / m   
    
    return b

BlendConcaveDec = blendConcaveDec


def blendConcaveCombined(d=0.0, u1=0.5, u2=2.0, s1=1.0, s2=1.5, h=1.0):
    """
        blending function combined concave

        Combines the increasing and decreasing concave
        functions into one function.
        Proceeds from the increasing function, changes
        to the decreasing function when d > s2

        u1 for the increasing portion of the function
        u2 for decreasing
        s1 for increasing, s2 for decreasing

       returns blend
    """
    d = float(d)
    u1 = float(u1)
    u2 = float(u2)    
    s1 = float(s1)
    s2 = float(s2)
    
    m = (2 * s1 - u1 - d) 
    n = (u2 - 2 * s2 + d) 
    #print(m, n)
    if m == 0:
        return h
    elif n == 0: 
        return h
    
    if d >= s1 and d <= s2:
        return h
    elif d < s1:
        b = h * (s1 - u1) / m   
    elif d > s2:
        b = (h * (u2 - s2)) / n
    
    return b

BlendConcaveCombined = blendConcaveCombined

def blendCauchian(d=0.0, a=1.0, b=0.0, c=2.0, h=1.0):
    """
       blending function polynomial
       d = x value
       a = area under curve/uncertainty radius
       b = horizontal shift
       c = curvature of function/tuning scale factor
       h = height of the set

       returns blend
    """
    d = float(d)
    a = float(a)
    b = float(b)
    c = float(c)
    h = float(h)

    if a == 0:
        raise ValueError("Param a must be a non zero value.")

    return h / (1 + math.pow(abs((d-b)/a), 2*c))

BlendCauchian = blendCauchian


def blendCosine(d=0.0, a=0.166666667, b=0.0, h=1.0):
    """
       blending function polynomial
       d = x value
       a = area under curve/uncertainty radius
       b = horizontal shift
       h = height of the set

       returns blend
    """
    d = float(d)
    a = float(a)
    b = float(b)
    h = float(h)

    if b - 1.0/a <= d <= b + 1.0/a:
        return 0.5 * h * (1 + math.cos(a * math.pi * (d - b)))
    else:
        return 0.0

BlendCosine = blendCosine


def blendSinc(d, u=1.0):
    """
       blending function sinc
       d = delta x = x - horizontal shift
       u = uncertainty radius of xabs estimate error

       returns blend
    """
    u = float(u)

    if d == 0:
        return 1

    if u == 0:
        raise ValueError("Params cannot be zero.")

    mu = math.sin(d/u) / (d/u)

    return mu if mu >= 0 else 0

BlendSinc = blendSinc


def blendSpike(d=0.0, u=0.0, s=1.0):
    """
        blending function spike
        d - Delta x = xabs - xdr
        u - Horizontal offset
        s - Tuning scale factor

        returns blend
    """

    d = float(d)
    u = float(u)
    s = float(s)

    b = math.e**(-(abs(s * (d - u))))

    return b

BlendSpike = blendSpike


def blendSigmoidInc(d=0.0, u=0.0, s=1.0):
    """
        blending function increasing sigmoid
        d - Delta x = xabs - xdr
        u - Horizontal offset
        s - Tuning scale factor

        returns blend
    """
    d = float(d)
    u = float(u)
    s = float(s)

    b = 1.0/(1+math.e**(s*(u-d)))

    return b

BlendSigmoidInc = blendSigmoidInc


def blendSigmoidDec(d=0.0, u=0.0, s=1.0):
    """
        blending function decreasing sigmoid
        d - Delta x = xabs - xdr
        u - Horizontal offset
        s - Tuning scale factor

        returns blend
    """
    d = float(d)
    u = float(u)
    s = float(s)

    b = 1.0 - (1.0 / (1 + math.e ** (s * (u - d))))

    return b

BlendSigmoidDec = blendSigmoidDec

def blendTriangular(d, u=0.1, s=0.4, c=0.9):
    """
    Triangular blending funciton, taken from eq. 3.5
    
    c must be greater than s
    s must be greater than u
    u is the beginning point of the triangle
    c is the endpoint of the triangle
    s is the peak of the triangle
    """
    d = float(d)
    u = float(u)
    s = float(s)
    c = float(c)

    if (s - u) == 0:
        return 0
    if (c - s) == 0:
        return 0
    
    if d <= u: 
        b = 0.0
    elif d > u and d <= s:
        b = (d - u)/(s - u)
    elif d > s and d < c:
        b = (c - d)/(c - s)
    else: 
        b = 0.0
    return b

BlendTriangular = blendTriangular





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Runs all the example FloScripts


"""
import sys
import os

import ioflo.app.run

PLAN_DIR_PATH = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    , 'plan')

def getPlanFiles(planDirPath=PLAN_DIR_PATH):
    planFiles = []
    for fname in os.listdir(os.path.abspath(planDirPath)):
        root, ext = os.path.splitext(fname)
        if ext != '.flo' or root.startswith('__') or root.startswith('gps'):
            continue

        planFiles.append(os.path.abspath(os.path.join(planDirPath, fname)))
    return planFiles

def main():
    """ Run example scripts"""
    failedCount = 0
    plans = getPlanFiles()
    for plan in plans:
        failed = False
        name, ext = os.path.splitext(os.path.basename(plan))
        try:
            skeddar = ioflo.app.run.run(  name=name,
                                filepath=plan,
                                period=0.0625,
                                verbose=1,
                                real=False,)

            print("Plan {0}\n  Skeddar {1}\n".format(plan, skeddar.name))

            if not skeddar.built:
                failed = True
            else:
                for house in skeddar.houses:
                    failure = house.metas['failure'].value
                    if failure:
                        failed = True
                        print("**** Failed in House = {0}. "
                              "Failure = {1}.\n".format(house.name, failure))
                    else:
                        print("**** Succeeded in House = {0}.\n".format(house.name))

        except Exception as ex:
            print("Failed with Exception: {0}\n".format(ex))
            failed = True

        if failed:
            failedCount += 1

    print("{0} failed out of {1}.\n".format(failedCount, len(plans)))

if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-
"""
Unittests
"""
from __future__ import absolute_import, division, print_function

import sys

import unittest

import os

from binascii import hexlify

# Import ioflo libs
from ioflo.aid.sixing import *
from ioflo.aid.byting import hexify, bytify, unbytify, packify, unpackify
from ioflo.aid import getConsole

console = getConsole()

from ioflo.aio.proto import packeting

def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    console.reinit(verbosity=console.Wordage.concise)

class BasicTestCase(unittest.TestCase):
    """
    Test Case
    """

    def setUp(self):
        """

        """
        console.reinit(verbosity=console.Wordage.profuse)

    def tearDown(self):
        """

        """
        console.reinit(verbosity=console.Wordage.concise)

    def testPart(self):
        """
        Test Part class
        """
        console.terse("{0}\n".format(self.testPart.__doc__))

        part = packeting.Part()
        self.assertEqual(part.size, 0)
        self.assertEqual(part.size, len(part))
        self.assertEqual(part.packed, bytearray([]))

        part = packeting.Part(size=2)
        self.assertEqual(part.size, 2)
        self.assertEqual(part.size, len(part))
        self.assertEqual(part.packed, bytearray([0, 0]))

        packed = bytearray([0, 1, 2, 3])
        part = packeting.Part(size=2, packed=packed)
        self.assertEqual(part.size, 4)
        self.assertEqual(part.size, len(packed))
        self.assertEqual(part.size, len(part))
        self.assertEqual(part.packed, packed)

    def testPackerPart(self):
        """
        Test PackerPart class
        """
        console.terse("{0}\n".format(self.testPackerPart.__doc__))

        part = packeting.PackerPart()
        self.assertEqual(part.fmt, '!')
        self.assertTrue(part.packer)
        self.assertEqual(part.size, 0)
        self.assertEqual(part.size, len(part))
        self.assertEqual(part.size, part.packer.size)
        self.assertEqual(part.packed, bytearray([]))
        show = part.show()
        self.assertEqual(show, '    PackerPart: packed=0x\n')

        packed = part.pack()
        self.assertEqual(packed, part.packed)
        self.assertEqual(part.packed, bytearray([]))
        show = part.show()
        self.assertEqual(show, '    PackerPart: packed=0x\n')

        raw = bytearray()
        offset = part.parse(raw=raw)
        self.assertEqual(offset, part.size)
        self.assertEqual(offset, 0)
        self.assertEqual(part.packed, bytearray([]))
        show = part.show()
        self.assertEqual(show, '    PackerPart: packed=0x\n')

        raw = bytearray([7, 2, 1, 0])
        offset = part.parse(raw=raw)
        self.assertEqual(offset, part.size)
        self.assertEqual(offset, 0)
        self.assertEqual(part.packed, bytearray([]))
        show = part.show()
        self.assertEqual(show, '    PackerPart: packed=0x\n')

        part = packeting.PackerPart(fmt="!BBB")
        self.assertEqual(part.fmt, '!BBB')
        self.assertTrue(part.packer)
        self.assertEqual(part.size, 3)
        self.assertEqual(part.size, len(part))
        self.assertEqual(part.size, part.packer.size)
        self.assertEqual(part.packed, bytearray([0, 0, 0]))
        show = part.show()
        self.assertEqual(show, '    PackerPart: packed=0x000000\n')

    def testPackifierPart(self):
        """
        Test PackifierPart class
        """
        console.terse("{0}\n".format(self.testPackifierPart.__doc__))

        part = packeting.PackifierPart()
        self.assertEqual(part.fmt, '')
        self.assertEqual(part.size, 0)
        self.assertEqual(part.size, len(part))
        self.assertEqual(part.packed, bytearray([]))
        show = part.show()
        self.assertEqual(show, '    PackifierPart: packed=0x\n')

        packed = part.pack()
        self.assertEqual(packed, part.packed)
        self.assertEqual(part.packed, bytearray([]))
        show = part.show()
        self.assertEqual(show, '    PackifierPart: packed=0x\n')

        raw = bytearray([7, 2, 1, 0])
        offset = part.parse(raw=raw)
        self.assertEqual(offset, part.size)
        self.assertEqual(offset, 0)
        self.assertEqual(part.packed, bytearray([]))
        show = part.show()
        self.assertEqual(show, '    PackifierPart: packed=0x\n')

        part = packeting.PackifierPart(fmt="8 8 8")
        self.assertEqual(part.fmt, "8 8 8")
        self.assertEqual(part.fmtSize, 3)
        self.assertEqual(part.size, 3)
        self.assertEqual(part.size, len(part))
        self.assertEqual(part.fmtSize, part.size)
        self.assertEqual(part.packed, bytearray([0, 0, 0]))
        show = part.show()
        self.assertEqual(show, '    PackifierPart: packed=0x000000\n')

    def testPacketPart(self):
        """
        Test PacketPart class
        """
        console.terse("{0}\n".format(self.testPacketPart.__doc__))

        part = packeting.PacketPart()
        self.assertEqual(part.packet, None)
        self.assertEqual(part.size, 0)
        self.assertEqual(part.size, len(part))
        self.assertEqual(part.packed, bytearray([]))
        show = part.show()
        self.assertEqual(show, '    PacketPart: packed=0x\n')

        part = packeting.PacketPart(packet="Not a packet")
        self.assertEqual(part.packet, "Not a packet")


    def testPacket(self):
        """
        Test Packet class
        """
        console.terse("{0}\n".format(self.testPacket.__doc__))

        part = packeting.Packet()
        self.assertEqual(part.stack, None)
        self.assertEqual(part.size, 0)
        self.assertEqual(part.size, len(part))
        self.assertEqual(part.packed, bytearray([]))

        part = packeting.Packet(stack="Not a stack")
        self.assertEqual(part.stack, "Not a stack")

        packed = part.pack()
        self.assertEqual(packed, part.packed)
        self.assertEqual(part.packed, bytearray([]))

        raw = b'Hello'
        offset = part.parse(raw=raw)
        self.assertEqual(part.size, offset)
        self.assertEqual(part.packed, raw)


def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
             'testPart',
             'testPackerPart',
             'testPackifierPart',
             'testPacketPart',
             'testPacket',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testPart')






"""needing.py need action module

"""
#print("module {0}".format(__name__))

import time
import struct
from collections import deque
import inspect

from ..aid.sixing import *
from .globaling import *
from ..aid.odicting import odict
from ..aid import aiding
from ..aid.timing import tuuid
from . import excepting
from . import registering
from . import storing
from . import acting
from . import tasking
from . import framing

from ..aid.consoling import getConsole
console = getConsole()

class Need(acting.Actor):
    """Need Class for conditions  such as entry or trans
    """
    Registry = odict()

    def __init__(self, **kwa):
        """
        Initialize Instance

        Inherited Attributes:
            .name = name string for Actor variant in class Registry
            .store = reference to shared data Store
            ._act = reference to containing Act

        Attributes:
            ._tracts = list of references to transition acts for this need
                transit sub-context of precur context during segue

        """
        super(Need, self).__init__(**kwa)
        self._tracts = []

    def _expose(self):
        """
        """
        print("Need %s " % (self.name))

    @staticmethod
    def Check(state, comparison, goal, tolerance):
        """Check state compared to goal with tolerance
           tolerance ignored unless comparison == or !=
        """
        if comparison == '==':
            try: #in case goal is string
                result = ( (goal - abs(tolerance)) <= state <= (goal + abs(tolerance)))
            except TypeError:
                result = (goal == state)
        elif comparison == '<':
            result = ( state < goal)
        elif comparison == '<=':
            result = ( state <= goal)
        elif comparison == '>=':
            result = ( state >= goal)
        elif comparison == '>':
            result = ( state > goal)
        elif comparison == '!=':
            try: #in case goal is string
                result = not ( (goal - abs(tolerance)) <= state <= (goal + abs(tolerance)) )
            except TypeError:
                result = (goal != state)
        else:
            result = False

        return result

    def addTract(self, act):
        """
        Add act to ._tracts list
        """
        self._tracts.append(act)
        act.frame = self._act.frame.name  # re-resolve later
        act.context = ActionSubContextNames[TRANSIT]


class NeedAlways(Need):
    """NeedAlways Need  Special Need"""
    def action(self, **kw):
        """Always return true"""
        result = True
        console.profuse("Need Always = {0}\n".format(result))
        return result

class NeedDone(Need):
    """
    NeedDone Need Special Need
    """
    def _resolve(self, tasker, **kwa):
        """Resolves value (tasker) link that is passed in as tasker parm
           resolved link is passed back to container ._act to update in act's parms
        """
        parms = super(NeedDone, self)._resolve( **kwa)

        parms['tasker'] = tasker = tasking.resolveTasker(tasker,
                                                        who=self._act.frame.name,
                                                        desc='need done',
                                                        contexts=[],
                                                        human=self._act.human,
                                                        count=self._act.count)

        return parms #return items are updated in original ._act parms

    def action(self, tasker, **kw):
        """
        Check if  tasker done
        parameters:
            tasker
        """
        result = tasker.done
        console.profuse("Need Tasker {0} done = {1}\n".format(tasker.name, result))
        return result

class NeedDoneAux(Need):
    """
    NeedDoneAux Need Special Need
    """
    def _resolve(self, tasker, framer, frame, **kwa):
        """
        Resolves aux framer tag link that is passed in as tasker parm relative to
        containing framer and frame

        Resolved link is passed back to container ._act to update in act's parms
        """
        parms = super(NeedDoneAux, self)._resolve( **kwa)
        if framer:
            if framer == 'me':
                framer = self._act.frame.framer
            parms['framer'] = framer = framing.resolveFramer(framer,
                                                            who=self._act.frame.name,
                                                            desc='need done',
                                                            contexts=[],
                                                            human=self._act.human,
                                                            count=self._act.count)

        if frame: # framer required
            if frame == 'me':
                frame = self._act.frame
            parms['frame'] = frame = framing.resolveFrameOfFramer(frame,
                                                                  framer,
                                                                  who=self._act.frame.name,
                                                                  desc='need done',
                                                                  human=self._act.human,
                                                                  count=self._act.count)

        if tasker not in ['any', 'all']:
            parms['tasker'] = tasker = framing.resolveAuxOfFramer(tasker,
                                                            framer,
                                                            who=self._act.frame.name,
                                                            desc='need done aux',
                                                            contexts=[],
                                                            human=self._act.human,
                                                            count=self._act.count)

        return parms #return items are updated in original ._act parms

    def action(self, tasker, framer, frame, **kw):
        """
        Check if  aux done
        parameters:
            aux
            framer
            frame
        """
        if frame:
            if tasker == 'any':
                result = any([aux.done for aux in frame.auxes ])
            elif tasker == 'all':
                result = frame.auxes and all([aux.done for aux in frame.auxes])
            elif tasker in frame.auxes:
                result = tasker.done
            else:
                result = False
            name = tasker if tasker in ('any', 'all') else tasker.tag
            console.profuse("Need Aux {0} done = {1} in {2}<{3}\n".format(name,
                                                                          result,
                                                                          framer.name,
                                                                          frame.name))
        else:
            result = tasker.done
            console.profuse("Need Aux {0} done = {1}\n".format(tasker.name, result))

        return result

class NeedStatus(Need):
    """NeedStatus Need Special Need """

    def _resolve(self, tasker, **kwa):
        """Resolves value (tasker) link that is passed in as parm
           resolved link is passed back to container ._act to update in act's parms
        """
        parms = super(NeedStatus, self)._resolve( **kwa)
        if tasker == 'me':
            tasker = self._act.frame.framer

        parms['tasker'] = tasker = tasking.resolveTasker(tasker,
                                                         who=self.name,
                                                         desc='need status tasker',
                                                         contexts=[],
                                                         human=self._act.human,
                                                         count=self._act.count)
        return parms #return items are updated in original ._act parms

    def action(self, tasker, status, **kw):
        """
        Check if  tasker done
        parameters:
          tasker
          status
        """
        # maybe should add check for auxiliary since status never changes for auxiliary

        result = (tasker.status == status)
        console.profuse("Need Tasker {0} status is {1} = {2}\n".format(
            tasker.name, StatusNames[status], result))

        return result


class NeedState(Need):
    """
    NeedState is a base class for Needs that must resolve a state share ref
    """
    def _resolve(self, state, stateField, **kwa):
        """
        Resolves state share

        parms:
            state       share path of state
            stateField  state share field name
        """
        parms = super(NeedState, self)._resolve( **kwa)

        #convert state path to share and create field if necessary
        parms['state'] = state = self._resolvePath(ipath=state,
                                                  warn=True) # now a share

        if not stateField: #default rules for field
            if state: #state has fields
                if 'value' in state:
                    stateField = 'value'

                else: #ambiguous
                    msg = ("ResolveError: Can't determine field for state"
                          " '{0}'".format(state.name))
                    raise excepting.ResolveError(msg, 'state', self.name,
                                            self._act.human, self._act.count)
            else:
                stateField = 'value'

        if stateField not in state:
            console.profuse("     Warning: Non-existent field '{0}' in state {1}"
                            " ... creating anyway".format(stateField, state.name))
            state[stateField] = 0.0 #create

        parms['stateField'] = stateField

        return parms #return items are updated in original ._act parms

class NeedBoolean(NeedState):
    """NeedBoolean Need Special Need

       if state
    """

    def action(self, state, stateField, **kw):
        """ Check if state[stateField] evaluates to True
            parameters:
              state = share of state
              stateField = field key

        """
        if state[stateField]:
            result = True
        else:
            result = False
        console.profuse("Need Boolean, if {0}[{1}]: = {2}\n".format(
            state.name, stateField, result))

        return result

class NeedDirect(NeedState):
    """NeedDirect Need

       if state comparison goal [+- tolerance]
    """

    def action(self, state, stateField, comparison, goal, tolerance, **kw):
        """ Check if state[field] comparison to goal +- tolerance is True
            parameters:
                state = share of state
                stateField = field key
                comparison
                goal
                tolerance

        """
        result = self.Check(state[stateField], comparison, goal, tolerance)
        console.profuse("Need Direct, if {0}[{1}] {2} {3} +- {4}: = {5}\n".format(
            state.name, stateField, comparison, goal, tolerance, result))

        return result

class NeedIndirect(NeedState):
    """NeedIndirect Need

       if state comparison goal [+- tolerance]
    """
    def _resolve(self, goal, goalField, **kwa):
        """
        Resolves state share

        parms:
            comparison  comparison operator string
            goal        share path of goal
            goalField   goal share field name
            tolerance   tolerance value

        """
        parms = super(NeedIndirect, self)._resolve( **kwa)

        #convert goal path to share and create field if necessary
        parms['goal'] = goal = self._resolvePath(ipath=goal,
                                                warn=True) # now a share

        if not goalField: #default rules for field
            if goal: #goal has fields
                if 'value' in goal:
                    goalField = 'value'

                else: #use stateField
                    goalField = stateField
            else:
                goalField = 'value'

        if goalField not in goal:
            console.profuse("     Warning: Non-existent field '{0}' in goal"
                    " {1} ... creating anyway".format(goalField, goal.name))
            goal[goalField] = 0.0 #create

        parms['goalField'] = goalField

        return parms #return items are updated in original ._act parms

    def action(self, state, stateField, comparison, goal, goalField, tolerance, **kwa):
        """ Check if state[field] comparison to goal[goalField] +- tolerance is True
                       parameters:
              state = share of state
              stateField = field key
              comparison
              goal
              goalField
              tolerance

        """

        result = self.Check(state[stateField], comparison, goal[goalField], tolerance)
        console.profuse("Need Indirect, if {0}[{1}] {2} {3}[{4}] +- %s: = {5}\n".format(
            state.name, stateField, comparison, goal, goalField, tolerance, result))

        return result

class NeedMarker(Need):
    """
    NeedMarker is base class for needs that insert markers on resolvelinks
    Special Need

    """
    def _resolve(self, share, frame, kind, marker, **kwa):
        """
        Resolves frame name link and then
           inserts marker as first enact in the resolved frame

        Incoming Parameters:
            share is path or ref of share holding mark
            frame is name or ref of frame to add option enact marker
                If frame empty then only put marker in transit sub context of
                need act's frame
            kind is name of marker actor class for marker act
            marker is unique identifier of marker if not empty
                This allows multiple if updated/changed to use same marker

        Outgoing Parameters:
            share is resolved ref of share holding mark
            marker is unique key for Mark instance in .marks odict in share

        Marker Act Parameters:
            share is resolved ref of share holding mark
            marker is unique key for Mark instance in .marks odict in share

        """
        parms = super(NeedMarker, self)._resolve( **kwa)

        framer = self._act.frame.framer
        enacted = True if frame else False

        if not frame or frame == 'me':
            frame = self._act.frame
        frame = framing.resolveFrameOfFramer(frame,
                                                framer,
                                                who=self.name,
                                                desc='need marker',
                                                human=self._act.human,
                                                count=self._act.count)

        parms['share'] = share = self._resolvePath(ipath=share,
                                                  warn=True) # now a share
        parts = [framer.name]
        if marker:
            parts.append(marker)  # framername.marker
        else:
            parts.append(frame.name)  # default is framername.framename

        marker = "<".join(parts)
        parms['marker'] = marker

        if not share.marks.get(marker):
            share.marks[marker] = storing.Mark()

        if kind not in acting.Actor.Registry:
            msg = "ResolveError: Bad need marker link"
            raise excepting.ResolveError(msg, kind, self.name,
                            self._act.human, self._act.count)

        markerParms = dict(share=share, marker=marker)
        markerAct = acting.Act(actor=kind,
                                registrar=acting.Actor,
                                parms=markerParms,
                                human=self._act.human,
                                count=self._act.count)

        self.addTract(markerAct)  # sets act.context to 'transit'
        console.profuse("     Added {0} {1} with {2} at {3} in {4} of "
                        "framer {5}\n".format(
                                'tract',
                                markerAct,
                                markerAct.parms['share'].name,
                                markerAct.parms['marker'],
                                self._act.frame.name,
                                framer.name))
        markerAct.resolve()

        if enacted:  # only add enact marker if original provided frame not empty
            found = False
            for enact in frame.enacts:  # avoid adding redundant marker
                if (isinstance(enact.actor, acting.Actor) and
                        enact.actor.name == kind and
                        enact.parms['share'].name == share.name and
                        enact.parms['marker'] == marker):
                    found = True
                    break

            if not found:
                markerParms = dict(share=share, marker=marker)
                markerAct = acting.Act(actor=kind,
                                                         registrar=acting.Actor,
                                                         parms=markerParms,
                                                         human=self._act.human,
                                                         count=self._act.count)

                frame.insertEnact(markerAct)
                console.profuse("     Added {0} {1} with {2} at {3} in {4} of "
                                "framer {5}\n".format(
                                        'enact',
                                        markerAct,
                                        markerAct.parms['share'].name,
                                        markerAct.parms['marker'],
                                        frame.name,
                                        framer.name))
                markerAct.resolve()  # resolves .actor given by actor kind name into actor class

        return parms #return items are updated in original ._act parms

class NeedUpdate(NeedMarker):
    """ NeedUpdate Need Special Need """
    def action(self, share, marker, **kw):
        """
        Check if share updated since mark in share was updated by marker
        Default is False

        Parameters:
            share is resolved share that is marked with .mark[marker]
            marker is marker key
        """
        result = False
        mark = share.marks.get(marker)
        if mark and share.stamp is not None:  # only if share updated from None
            # share stamps only non-None if updated after starts running not init
            # == catches updates to share on same enter or precur as Marker reset
            # mark.used only lets == work the first time
            # mark.stamp is None and share.stamp not None always True so if mark
            # not yet set will always work first time
            # store stamp is None until first run by skedder to share.stamp is
            # not updated until some action updates it. (not by init)
            result = ((mark.stamp is None) or
                      (share.stamp > mark.stamp) or
                      (share.stamp == mark.stamp and mark.used != mark.stamp))

        console.profuse("Marker update {0} for {1} of Share {2} {3} "
                        " {4} mark {5} used {6} at {7}\n".format(result,
                                                     marker,
                                                     share.name,
                                                     share.stamp,
                                                     '>=',
                                                     mark.stamp,
                                                     mark.used,
                                                     self.store.stamp))

        return result

class NeedChange(NeedMarker):
    """NeedChange Need Special Need"""
    def action(self, share, marker, **kw):
        """
        Check if share data changed while denoted by marker key if any
        Default is False

        Parameters:
            share is resolved share that is marked with .mark[marker]
            marker is marker key
        """
        result = False
        mark = share.marks.get(marker) #get mark from mark frame name key
        if mark:
            if mark.data is None:
                result = True  # always true first time if mark not yet set
            else:
                for field, value in share.items():
                    try:
                        if getattr(mark.data, field) != value:
                            result = True

                    except AttributeError as ex: # new attribute so changed
                        result = True  # should not remove fields so only test add case

                    if result: #stop checking on first change
                        break


        console.profuse("Marker change {0} for {1} of data {2} of share {3} at {4}\n".format(
            result, marker, mark.data if mark else None, share.name, self.store.stamp))

        return result
